# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from trytond.report import Report
from trytond.pool import Pool
from trytond.i18n import gettext

__all__ = ['PediatricsGrowthChartsWHOReport', 'WeightForAge',
           'LengthHeightForAge', 'BMIForAge']


class PediatricsGrowthChartsWHOReport(Report):
    __name__ = 'gnuhealth.pediatrics.growth.charts.who.report'

    @classmethod
    def get_context(cls, records, header, data):

        _MODULE = "health_pediatrics_growth_charts_who"

        _TYPES = {
            '-3': 'p3',
            '-2': 'p15',
            '0': 'p50',
            '2': 'p85',
            '3': 'p97',
        }

        _INDICATORS = {
            'l/h-f-a': gettext(
                _MODULE + ".msg_indicator_length_or_height_for_age"),
            'w-f-a': gettext(
                _MODULE + ".msg_indicator_weight_for_age"),
            'bmi-f-a': gettext(
                _MODULE + ".msg_indicator_bmi_for_age"),
        }

        _MEASURES = {
            'p': gettext(_MODULE + ".msg_measure_percentiles"),
            'z': gettext(_MODULE + ".msg_measure_z_scores"),
        }

        _GENDERS = {
            'f': gettext(_MODULE + ".msg_gender_girls"),
            'm': gettext(_MODULE + ".msg_gender_boys"),
        }

        _SUBTITLE = gettext(_MODULE + ".msg_subtitle")

        pool = Pool()
        GrowthChartsWHO = pool.get(
            'gnuhealth.pediatrics.growth.charts.who')
        Patient = pool.get('gnuhealth.patient')
        Evaluation = pool.get('gnuhealth.patient.evaluation')

        context = super(
            PediatricsGrowthChartsWHOReport, cls).get_context(
                records, header, data)

        patient = Patient(data['patient'])

        growthchartswho = GrowthChartsWHO.search([
            ('indicator', '=', data['indicator']),
            ('measure', '=', data['measure']),
            ('sex', '=', patient.party.gender),
        ], order=[('month', 'ASC')],
        )

        context['title'] = _INDICATORS[data['indicator']].format(
            gender=_GENDERS[patient.party.gender])
        context['subtitle'] = _SUBTITLE.format(
            measure=_MEASURES[data['measure']])
        context['name'] = patient.party.rec_name
        context['puid'] = patient.puid
        context['date'] = datetime.now().date()
        context['age'] = patient.age
        context['measure'] = data['measure']

        if data['measure'] == 'p':
            context['p3'] = gettext(
                _MODULE + ".msg_table_header_percentile_p3"),
            context['p15'] = gettext(
                _MODULE + ".msg_table_header_percentile_p15"),
            context['p50'] = gettext(
                _MODULE + ".msg_table_header_percentile_p50"),
            context['p85'] = gettext(
                _MODULE + ".msg_table_header_percentile_p85"),
            context['p97'] = gettext(
                _MODULE + ".msg_table_header_percentile_p97"),
        else:
            context['p3'] = gettext(
                _MODULE + ".msg_table_header_z_score_p3"),
            context['p15'] = gettext(
                _MODULE + ".msg_table_header_z_score_p15"),
            context['p50'] = gettext(
                _MODULE + ".msg_table_header_z_score_p50"),
            context['p85'] = gettext(
                _MODULE + ".msg_table_header_z_score_p85"),
            context['p97'] = gettext(
                _MODULE + ".msg_table_header_z_score_p97"),

        for value in growthchartswho:
            if data['measure'] == 'p':
                context[value.type.lower() + '_' + str(value.month)] = \
                    value.value
            else:
                context[_TYPES[value.type] + '_' + str(value.month)] = \
                    value.value

        evaluations = Evaluation.search([
            ('patient', '=', data['patient']),
        ])

        for month in range(61):
            context['v' + str(month)] = ''

        for evaluation in evaluations:
            if evaluation.age_months is not None:
                con = ''.join(['v', str(evaluation.age_months)])
                if data['indicator'] == 'l/h-f-a':
                    context[con] = evaluation.height
                elif data['indicator'] == 'w-f-a':
                    context[con] = evaluation.weight
                else:
                    context[con] = evaluation.bmi

        return context


class WeightForAge(PediatricsGrowthChartsWHOReport):
    __name__ = 'gnuhealth.pediatrics.growth.charts.who.wfa.report'


class LengthHeightForAge(PediatricsGrowthChartsWHOReport):
    __name__ = 'gnuhealth.pediatrics.growth.charts.who.lhfa.report'


class BMIForAge(PediatricsGrowthChartsWHOReport):
    __name__ = 'gnuhealth.pediatrics.growth.charts.who.bmifa.report'



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                      HEALTH ISS package                               #
#                        health_iss.py                                  #
#########################################################################
from trytond.pyson import Eval, Not, Equal
from trytond.model import ModelView, ModelSQL, fields, Unique

__all__ = ['Iss']


class Iss (ModelSQL, ModelView):
    'Injury Surveillance System Registration'
    __name__ = 'gnuhealth.iss'

    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation',
        'Evaluation', required=True, help='Related Patient Evaluation')

    injury_date = fields.Date(
        'Injury Date',
        help="Usually the same as the Evaluation")

    registration_date = fields.Date('Registration Date')

    code = fields.Char('Code', help='Injury Code', required=True)

    operational_sector = fields.Many2One(
        'gnuhealth.operational_sector',
        'O. Sector', help="Operational Sector in which happened the injury")

    latitude = fields.Numeric('Latidude', digits=(3, 14))
    longitude = fields.Numeric('Longitude', digits=(4, 14))

    urladdr = fields.Char(
        'OSM Map',
        help="Maps the Accident / Injury location on Open Street Map")

    healthcenter = fields.Many2One('gnuhealth.institution', 'Institution')

    patient = fields.Function(
        fields.Char('Patient'),
        'get_patient', searcher='search_patient')

    patient_sex = fields.Function(
        fields.Char('Gender'),
        'get_patient_sex')

    patient_age = fields.Function(fields.Char('Age'), 'get_patient_age')

    complaint = fields.Function(
        fields.Char('Chief Complaint'),
        'get_patient_complaint')

    injury_type = fields.Selection([
        (None, ''),
        ('accidental', 'Accidental / Unintentional'),
        ('violence', 'Violence'),
        ('attempt_suicide', 'Suicide Attempt'),
        ('motor_vehicle', 'Motor Vehicle'),
    ], 'Injury Type', required=True, sort=False)

    mva_mode = fields.Selection([
        (None, ''),
        ('pedestrian', 'Pedestrian'),
        ('bicycle', 'Bicycle'),
        ('motorbike', 'Motorbike'),
        ('car', 'Car'),
        ('van', 'Van / Pickup / Jeep'),
        ('truck', 'Truck / Heavy vehicle'),
        ('bus', 'Bus'),
        ('train', 'Train'),
        ('taxi', 'Taxi'),
        ('boat', 'Boat / Ship'),
        ('aircraft', 'Aircraft'),
        ('other', 'Other'),
        ('unknown', 'Unknown'),
    ], 'Mode', help="Motor Vehicle Accident Mode", sort=False,
        states={'required': Equal(Eval('injury_type'), 'motor_vehicle')})

    mva_position = fields.Selection([
        (None, ''),
        ('driver', 'Driver'),
        ('passenger', 'Passenger'),
        ('outside', 'Outside / on the back'),
        ('bystander', 'Bystander'),
        ('unspecified_vehicle', 'Unspecified vehicle'),
        ('unknown', 'Unknown'),
    ], 'User Position',
        help="Motor Vehicle Accident user position", sort=False,
        states={'required': Equal(Eval('injury_type'), 'motor_vehicle')})

    mva_counterpart = fields.Selection([
        (None, ''),
        ('pedestrian', 'Pedestrian'),
        ('bicycle', 'Bicycle'),
        ('motorbike', 'Motorbike'),
        ('car', 'Car'),
        ('van', 'Van / Pickup / Jeep'),
        ('truck', 'Truck / Heavy vehicle'),
        ('bus', 'Bus'),
        ('train', 'Train'),
        ('taxi', 'Taxi'),
        ('boat', 'Boat / Ship'),
        ('aircraft', 'Aircraft'),
        ('other', 'Other'),
        ('unknown', 'Unknown'),
    ], 'Counterpart',
        help="Motor Vehicle Accident Counterpart", sort=False,
        states={'required': Equal(Eval('injury_type'), 'motor_vehicle')})

    safety_gear = fields.Selection([
        (None, ''),
        ('yes', 'Yes'),
        ('no', 'No'),
        ('unknown', 'Unknown'),
    ], 'Safety Gear',
        help="Use of Safety Gear - Helmet, safety belt...", sort=False,
        states={'required': Equal(Eval('injury_type'), 'motor_vehicle')})

    alcohol = fields.Selection([
        (None, ''),
        ('yes', 'Yes'),
        ('no', 'No'),
        ('suspected', 'Suspected'),
        ('unknown', 'Unknown'),
    ], 'Alcohol', required=True,
        help="Is there evidence of alcohol use by the injured person"
        " in the 6 hours before the accident ?", sort=False)

    drugs = fields.Selection([
        (None, ''),
        ('yes', 'Yes'),
        ('no', 'No'),
        ('suspected', 'Suspected'),
        ('unknown', 'Unknown'),
    ], 'Other Drugs', required=True,
        help="Is there evidence of drug use by the injured person"
        " in the 6 hours before the accident ?", sort=False)

    injury_details = fields.Text('Details')

    # Add victim-perpretator relationship for violence-related injuries
    victim_perpetrator = fields.Selection([
        (None, ''),
        ('parent', 'Parent'),
        ('spouse', 'Wife / Husband'),
        ('girlboyfriend', 'Girl / Boyfriend'),
        ('relative', 'Other relative'),
        ('acquaintance', 'Acquaintance / Friend'),
        ('official', 'Official / Legal'),
        ('stranger', 'Stranger'),
        ('other', 'other'),
    ], 'Relationship',
        help="Victim - Perpetrator relationship", sort=False,
        states={'required': Equal(Eval('injury_type'), 'violence')})

    violence_circumstances = fields.Selection([
        (None, ''),
        ('fight', 'Fight'),
        ('robbery', 'Robbery'),
        ('drug', 'Drug Related'),
        ('sexual', 'Sexual Assault'),
        ('gang', 'Gang Activity'),
        ('other_crime', 'Committing a crime (other)'),
        ('other', 'Other'),
        ('unknown', 'Unknown'),
    ], 'Context',
        help="Precipitating Factor", sort=False,
        states={'required': Equal(Eval('injury_type'), 'violence')})

    injury_method = fields.Selection([
        (None, ''),
        ('blunt', 'Blunt object'),
        ('push', 'Push/bodily force'),
        ('sharp', 'Sharp objects'),
        ('gun', 'Gun shot'),
        ('sexual', 'Sexual Assault'),
        ('choking', 'Choking/strangulation'),
        ('other', 'Other'),
        ('unknown', 'Unknown'),
    ], 'Method',
        help="Method of Injury", sort=False,
        states={'required': Equal(Eval('injury_type'), 'violence')})

    # Place of occurrance . Not used in motor vehicle accidents

    place_occurrance = fields.Selection([
        (None, ''),
        ('home', 'Home'),
        ('street', 'Street'),
        ('institution', 'Institution'),
        ('school', 'School'),
        ('commerce', 'Commercial Area'),
        ('publicbuilding', 'Public Building'),
        ('recreational', 'Recreational Area'),
        ('transportation', 'Public transportation'),
        ('sports', 'Sports event'),
        ('unknown', 'Unknown'),
    ], 'Place',
        help="Place of occurrance", sort=False,
        states={
        'required': Not(Equal(Eval('injury_type'), 'motor_vehicle'))})

    disposition = fields.Selection([
        (None, ''),
        ('treated_sent', 'Treated and Sent Home'),
        ('admitted', 'Admitted to Ward'),
        ('observation', 'Admitted to Observation'),
        ('died', 'Died'),
        ('daa', 'Discharge Against Advise'),
        ('transferred', 'Transferred'),
        ('doa', 'Dead on Arrival'),
    ], 'Disposition',
        help="Place of occurrance", sort=False, required=True)

    def get_patient(self, name):
        return self.evaluation.patient.rec_name

    def get_patient_sex(self, name):
        return self.evaluation.patient.gender_str

    def get_patient_age(self, name):
        return self.evaluation.patient.age

    def get_patient_complaint(self, name):
        return self.evaluation.chief_complaint

    @fields.depends('latitude', 'longitude')
    def on_change_with_urladdr(self):
        # Generates the URL to be used in OpenStreetMap
        # The address will be mapped to the URL in the following way
        # If the latitud and longitude of the Accident / Injury
        # are given, then those parameters will be used.

        ret_url = ''
        if (self.latitude and self.longitude):
            ret_url = 'http://openstreetmap.org/?mlat=' + \
                str(self.latitude) + '&mlon=' + str(self.longitude)

        return ret_url

    @classmethod
    def search_patient(cls, name, clause):
        res = []
        value = clause[2]
        res.append(('evaluation.patient', clause[1], value))
        return res

    @classmethod
    def view_attributes(cls):
        return [('//group[@id="motor_vehicle_accident"]', 'states', {
                'invisible': Not(Equal(Eval('injury_type'), 'motor_vehicle')),
                }),
                ('//group[@id="violent_injury"]', 'states', {
                    'invisible': Not(Equal(Eval('injury_type'), 'violence')),
                }),
                ('//group[@id="iss_place"]', 'states', {
                    'invisible': Equal(Eval('injury_type'), 'motor_vehicle'),
                }),
                ]

    @classmethod
    def __setup__(cls):
        super(Iss, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('code_uniq', Unique(t, t.code),
             'This ISS registration Code already exists'),
        ]

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to evaluation
        if (table_h.column_exist('name')
                and not table_h.column_exist('evaluation')):
            table_h.column_rename('name', 'evaluation')

        super().__register__(module)
        table_h = cls.__table_handler__(module)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                  HEALTH STOCK SURGERY package                         #
#              health_stock_surgery.py: main module                     #
#########################################################################

from trytond.model import fields

from trytond.pool import PoolMeta

__all__ = ['Move', 'Surgery']


class Move(metaclass=PoolMeta):
    __name__ = 'stock.move'

    @classmethod
    def _get_origin(cls):
        return super(Move, cls)._get_origin() + [
            'gnuhealth.surgery',
        ]


class Surgery(metaclass=PoolMeta):
    __name__ = 'gnuhealth.surgery'

    moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
    location = fields.Many2One(
        'stock.location',
        'Stock Location', domain=[('type', '=', 'storage')])



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2014 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import wizard_top_diseases
from . import wizard_evaluations
from . import wizard_summary_report
from . import wizard_epidemics_report


__all__ = ['wizard_top_diseases', 'wizard_evaluations',
           'wizard_summary_report', 'wizard_epidemics_report']



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#                test_health.py health unittest file                    #
#########################################################################
import unittest
import trytond.tests.test_tryton
from trytond.tests.test_tryton import ModuleTestCase


class HealthTestCase(ModuleTestCase):
    "Test Health package"
    module = 'health'


del ModuleTestCase


def suite():
    suite = trytond.tests.test_tryton.suite()
    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(
        HealthTestCase))
    return suite



# SPDX-FileCopyrightText: 2009-2013 Bertrand Chenal
# SPDX-FileCopyrightText: 2009-2016 B2CK
# SPDX-FileCopyrightText: 2009-2016 Cédric Krier
# SPDX-FileCopyrightText: 2009-2016 Tryton Foundation <info@tryton.org>
# SPDX-FileCopyrightText: 2016-2024 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2016-2024 Luis Falcón <falcon@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import uuid
import vobject
import dateutil.tz
import pytz
import datetime
import xml.dom.minidom
from sql import Null

from trytond.model import Model, ModelSQL, ModelView, fields, Check, Unique
from trytond.tools import reduce_ids, grouped_slice
from trytond.pyson import If, Bool, Eval
from trytond.transaction import Transaction
from trytond.cache import Cache
from trytond.pool import Pool
from trytond.i18n import gettext

from .exceptions import (
    InvalidCalendarExtension, InvalidRecurrence,
    InvalidBySecond, InvalidByMinute,
    InvalidByHour, InvalidByDay,
    InvalidByMonthDay, InvalidByWeekNumber,
    InvalidByYearDay,
    InvalidByMonth, InvalidBySetPosition)

__all__ = ['Calendar', 'ReadUser', 'WriteUser', 'Category', 'Location',
           'Event', 'EventCategory', 'AlarmMixin', 'EventAlarm',
           'AttendeeMixin',
           'EventAttendee', 'DateMixin', 'EventRDate', 'EventExDate',
           'RRuleMixin',
           'EventRRule', 'EventExRule']


tzlocal = dateutil.tz.tzlocal()
tzutc = dateutil.tz.tzutc()
domimpl = xml.dom.minidom.getDOMImplementation()


class Calendar(ModelSQL, ModelView):
    "Calendar"
    __name__ = 'calendar.calendar'
    name = fields.Char('Name', required=True)
    description = fields.Text('Description')
    owner = fields.Many2One(
        'res.user', 'Owner',
        domain=[('email', '!=', None)],
        help='The user must have an email')
    read_users = fields.Many2Many(
        'calendar.calendar-read-res.user',
        'calendar', 'user', 'Read Users')
    write_users = fields.Many2Many(
        'calendar.calendar-write-res.user',
        'calendar', 'user', 'Write Users')
    _get_name_cache = Cache('calendar_calendar.get_name')

    @classmethod
    def __setup__(cls):
        super(Calendar, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('name_uniq', Unique(t, t.name),
                'The name of calendar must be unique.'),
            ('owner_uniq', Unique(t, t.owner),
                'A user can have only one calendar.'),
        ]
        cls._order.insert(0, ('name', 'ASC'))

    @classmethod
    def create(cls, vlist):
        calendars = super(Calendar, cls).create(vlist)
        # Restart the cache for get_name
        cls._get_name_cache.clear()
        return calendars

    @classmethod
    def write(cls, calendars, values, *args):
        super(Calendar, cls).write(calendars, values, *args)
        # Restart the cache for get_name
        cls._get_name_cache.clear()

    @classmethod
    def delete(cls, calendars):
        super(Calendar, cls).delete(calendars)
        # Restart the cache for calendar
        cls._get_name_cache.clear()

    @classmethod
    def validate(cls, calendars):
        super(Calendar, cls).validate(calendars)
        for calendar in calendars:
            calendar.check_name()

    def check_name(self):
        '''
        Check the name doesn't end with .ics
        '''
        if self.name.endswith('.ics'):
            raise InvalidCalendarExtension(
                gettext('health_caldav.msg_invalid_calendar_extension'))

    @classmethod
    def get_name(cls, name):
        '''
        Return the calendar id of the name
        '''
        calendar_id = cls._get_name_cache.get(name, default=-1)
        if calendar_id == -1:
            calendars = cls.search([
                ('name', '=', name),
            ], limit=1)
            if calendars:
                calendar_id = calendars[0].id
            else:
                calendar_id = None
            cls._get_name_cache.set(name, calendar_id)
        return calendar_id

    def calendar2ical(self):
        '''
        Return an iCalendar object for the given calendar_id containing
        all the vevent objects
        '''
        Event = Pool().get('calendar.event')

        ical = vobject.iCalendar()
        ical.vevent_list = []
        events = Event.search([
            ('calendar', '=', self.id),
            ('parent', '=', None),
        ])
        for event in events:
            ical2 = event.event2ical()
            ical.vevent_list.extend(ical2.vevent_list)
        return ical

    @classmethod
    def freebusy(cls, calendar_id, dtstart, dtend):
        '''
        Return an iCalendar object for the given calendar_id with the
        vfreebusy objects between the two dates
        '''
        Event = Pool().get('calendar.event')

        ical = vobject.iCalendar()
        ical.add('method').value = 'REPLY'
        ical.add('vfreebusy')
        if not isinstance(dtstart, datetime.datetime):
            ical.vfreebusy.add('dtstart').value = dtstart
            dtstart = datetime.datetime.combine(dtstart, datetime.time())\
                .replace(tzinfo=tzlocal)
        else:
            ical.vfreebusy.add('dtstart').value = dtstart.astimezone(tzutc)
        if not isinstance(dtend, datetime.datetime):
            ical.vfreebusy.add('dtend').value = dtend
            dtend = datetime.datetime.combine(dtend, datetime.time.max)\
                .replace(tzinfo=tzlocal)
        else:
            ical.vfreebusy.add('dtend').value = dtend.astimezone(tzutc)

        with Transaction().set_user(0):
            events = Event.search([
                ['OR',
                 [('dtstart', '<=', dtstart),
                  ('dtend', '>=', dtstart)],
                 [('dtstart', '<=', dtend),
                  ('dtend', '>=', dtend)],
                 [('dtstart', '>=', dtstart),
                  ('dtend', '<=', dtend)],
                 [('dtstart', '>=', dtstart),
                  ('dtstart', '<=', dtend),
                  ('dtend', '=', None)]],
                ('parent', '=', None),
                ('rdates', '=', None),
                ('rrules', '=', None),
                ('exdates', '=', None),
                ('exrules', '=', None),
                ('occurences', '=', None),
                ('calendar', '=', calendar_id),
            ])

        for event in events:
            # Don't group freebusy as sunbird doesn't handle it
            freebusy = ical.vfreebusy.add('freebusy')
            freebusy.fbtype_param = event._fbtype
            if event.dtstart.replace(tzinfo=tzlocal) >= dtstart:
                freebusy_dtstart = event.dtstart.replace(tzinfo=tzlocal)
            else:
                freebusy_dtstart = dtstart
            if event.dtend.replace(tzinfo=tzlocal) <= dtend:
                freebusy_dtend = event.dtend.replace(tzinfo=tzlocal)
            else:
                freebusy_dtend = dtend
            freebusy.value = [(
                freebusy_dtstart.astimezone(tzutc),
                freebusy_dtend.astimezone(tzutc))]

        with Transaction().set_user(0):
            events = Event.search([
                ('parent', '=', None),
                ('dtstart', '<=', dtend),
                ['OR',
                 ('rdates', '!=', None),
                 ('rrules', '!=', None),
                 ('exdates', '!=', None),
                 ('exrules', '!=', None),
                 ('occurences', '!=', None),
                 ],
                ('calendar', '=', calendar_id),
            ])

        for event in events:
            event_ical = event.event2ical()
            if event_ical.vevent.rruleset:
                for freebusy_dtstart in event_ical.vevent.rruleset:
                    if freebusy_dtstart.replace(tzinfo=tzlocal) > dtend:
                        break
                    if not event.dtend:
                        freebusy_dtend = freebusy_dtstart
                    else:
                        freebusy_dtend = event.dtend.replace(tzinfo=tzlocal)\
                            - event.dtstart.replace(tzinfo=tzlocal) \
                            + freebusy_dtstart
                    f_dtstart_tz = freebusy_dtstart.replace(tzinfo=tzlocal)
                    f_dtend_tz = freebusy_dtend.replace(tzinfo=tzlocal)
                    if not ((f_dtstart_tz <= dtstart
                             and f_dtend_tz >= dtstart)
                            or (f_dtstart_tz <= dtend
                                and f_dtend_tz >= dtend)
                            or (f_dtstart_tz >= dtstart
                                and f_dtend_tz <= dtend)):
                        continue
                    freebusy_fbtype = event._fbtype
                    all_day = event.all_day
                    for occurence in event.occurences:
                        if (occurence.recurrence.replace(tzinfo=tzlocal)
                                == f_dtstart_tz):
                            freebusy_dtstart = \
                                occurence.dtstart.replace(tzinfo=tzlocal)
                            if occurence.dtend:
                                freebusy_dtend = occurence.dtend\
                                    .replace(tzinfo=tzlocal)
                            else:
                                freebusy_dtend = freebusy_dtstart
                            all_day = occurence.all_day
                            freebusy_fbtype = occurence._fbtype
                            break
                    freebusy = ical.vfreebusy.add('freebusy')
                    freebusy.fbtype_param = freebusy_fbtype
                    if f_dtstart_tz <= dtstart:
                        freebusy_dtstart = dtstart
                    if f_dtend_tz >= dtend:
                        freebusy_dtend = dtend
                    if all_day:
                        freebusy.value = [(
                            f_dtstart_tz.astimezone(tzutc),
                            f_dtend_tz.astimezone(tzutc),
                        )]
                    else:
                        freebusy.value = [(
                            freebusy_dtstart.astimezone(tzutc),
                            freebusy_dtend.astimezone(tzutc))]
        return ical

    @classmethod
    def post(cls, uri, data):
        '''
        Handle post of vfreebusy request and return the XML with
        schedule-response
        '''
        from pywebdav.lib.errors import DAV_Forbidden
        Collection = Pool().get('webdav.collection')

        calendar_id = Collection.calendar(uri)
        if not calendar_id:
            raise DAV_Forbidden
        calendar = cls(calendar_id)
        if calendar.owner.id != Transaction().user:
            raise DAV_Forbidden
        ical = vobject.readOne(data)
        if ical.method.value == 'REQUEST' \
                and hasattr(ical, 'vfreebusy'):
            doc = domimpl.createDocument(None, 'schedule-response', None)
            sr = doc.documentElement
            sr.setAttribute('xmlns:D', 'DAV:')
            sr.setAttribute('xmlns:C', 'urn:ietf:params:xml:ns:caldav')
            sr.tagName = 'C:schedule-response'

            if not isinstance(ical.vfreebusy.dtstart.value, datetime.datetime):
                dtstart = ical.vfreebusy.dtstart.value
            else:
                if ical.vfreebusy.dtstart.value.tzinfo:
                    dtstart = ical.vfreebusy.dtstart.value.astimezone(tzlocal)
                else:
                    dtstart = ical.vfreebusy.dtstart.value
            if not isinstance(ical.vfreebusy.dtend.value, datetime.datetime):
                dtend = ical.vfreebusy.dtend.value
            else:
                if ical.vfreebusy.dtend.value.tzinfo:
                    dtend = ical.vfreebusy.dtend.value.astimezone(tzlocal)
                else:
                    dtend = ical.vfreebusy.dtend.value
            for attendee in ical.vfreebusy.attendee_list:
                resp = doc.createElement('C:response')
                sr.appendChild(resp)
                recipient = doc.createElement('C:recipient')
                href = doc.createElement('D:href')
                huri = doc.createTextNode(attendee.value)
                href.appendChild(huri)
                recipient.appendChild(href)
                resp.appendChild(recipient)

                vfreebusy = None
                email = attendee.value
                if attendee.value.lower().startswith('mailto:'):
                    email = attendee.value[7:]
                with Transaction().set_user(0):
                    calendars = cls.search([
                        ('owner.email', '=', email),
                    ])
                if calendars:
                    vfreebusy = cls.freebusy(calendars[0].id, dtstart, dtend)
                    vfreebusy.vfreebusy.add('dtstamp').value = \
                        ical.vfreebusy.dtstamp.value
                    vfreebusy.vfreebusy.add('uid').value = \
                        ical.vfreebusy.uid.value
                    vfreebusy.vfreebusy.add('organizer').value = \
                        ical.vfreebusy.organizer.value
                    vfreebusy.vfreebusy.add('attendee').value = attendee.value

                status = doc.createElement('C:request-status')
                status.appendChild(
                    doc.createTextNode(
                        vfreebusy and '2.0;Success'
                        or '5.3;No scheduling support for user.'))
                resp.appendChild(status)
                if vfreebusy:
                    data = doc.createElement('C:calendar-data')
                    data.appendChild(doc.createTextNode(vfreebusy.serialize()))
                    resp.appendChild(data)
            return doc.toxml(encoding='utf-8')
        raise DAV_Forbidden


class ReadUser(ModelSQL):
    'Calendar - read - User'
    __name__ = 'calendar.calendar-read-res.user'
    calendar = fields.Many2One(
        'calendar.calendar', 'Calendar',
        ondelete='CASCADE', required=True)
    user = fields.Many2One(
        'res.user', 'User', ondelete='CASCADE',
        required=True)


class WriteUser(ModelSQL):
    'Calendar - write - User'
    __name__ = 'calendar.calendar-write-res.user'
    calendar = fields.Many2One(
        'calendar.calendar', 'Calendar',
        ondelete='CASCADE', required=True)

    user = fields.Many2One(
        'res.user', 'User', ondelete='CASCADE',
        required=True)


class Category(ModelSQL, ModelView):
    "Category"
    __name__ = 'calendar.category'
    name = fields.Char('Name', required=True)

    @classmethod
    def __setup__(cls):
        super(Category, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('name_uniq', Unique(t, t.name),
                'The name of calendar category must be unique.'),
        ]
        cls._order.insert(0, ('name', 'ASC'))


class Location(ModelSQL, ModelView):
    "Location"
    __name__ = 'calendar.location'
    name = fields.Char('Name', required=True)

    @classmethod
    def __setup__(cls):
        super(Location, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('name_uniq', Unique(t, t.name),
                'The name of calendar location must be unique.'),
        ]
        cls._order.insert(0, ('name', 'ASC'))


class Event(ModelSQL, ModelView):
    "Event"
    __name__ = 'calendar.event'
    _rec_name = 'summary'
    uuid = fields.Char(
        'UUID', required=True,
        help='Universally Unique Identifier')
    calendar = fields.Many2One(
        'calendar.calendar', 'Calendar',
        required=True, ondelete="CASCADE")
    summary = fields.Char('Summary')
    sequence = fields.Integer('Sequence', required=True)
    description = fields.Text('Description')
    all_day = fields.Boolean('All Day')
    dtstart = fields.DateTime('Start Date', required=True)
    dtend = fields.DateTime('End Date')
    timezone = fields.Selection('timezones', 'Timezone')
    categories = fields.Many2Many(
        'calendar.event-calendar.category',
        'event', 'category', 'Categories')
    classification = fields.Selection([
        ('public', 'Public'),
        ('private', 'Private'),
        ('confidential', 'Confidential'),
    ], 'Classification', required=True)
    location = fields.Many2One('calendar.location', 'Location')
    status = fields.Selection([
        ('', ''),
        ('tentative', 'Tentative'),
        ('confirmed', 'Confirmed'),
        ('cancelled', 'Cancelled'),
    ], 'Status')
    organizer = fields.Char('Organizer', states={
        'required': If(Bool(Eval('attendees')), ~Eval('parent'), False),
    }, depends=['attendees', 'parent'])
    attendees = fields.One2Many(
        'calendar.event.attendee', 'event',
        'Attendees')
    transp = fields.Selection([
        ('opaque', 'Opaque'),
        ('transparent', 'Transparent'),
    ], 'Time Transparency', required=True)
    alarms = fields.One2Many('calendar.event.alarm', 'event', 'Alarms')
    rdates = fields.One2Many(
        'calendar.event.rdate', 'event',
        'Recurrence Dates',
        states={
            'invisible': Bool(Eval('parent')),
        }, depends=['parent'])
    rrules = fields.One2Many(
        'calendar.event.rrule', 'event',
        'Recurrence Rules',
        states={
            'invisible': Bool(Eval('parent')),
        }, depends=['parent'])
    exdates = fields.One2Many(
        'calendar.event.exdate', 'event',
        'Exception Dates',
        states={
            'invisible': Bool(Eval('parent')),
        }, depends=['parent'])
    exrules = fields.One2Many(
        'calendar.event.exrule', 'event',
        'Exception Rules',
        states={
            'invisible': Bool(Eval('parent')),
        }, depends=['parent'])
    occurences = fields.One2Many(
        'calendar.event', 'parent', 'Occurences',
        domain=[
            ('uuid', '=', Eval('uuid')),
            ('calendar', '=', Eval('calendar')),
        ],
        states={
            'invisible': Bool(Eval('parent')),
        }, depends=['uuid', 'calendar', 'parent'])
    parent = fields.Many2One(
        'calendar.event', 'Parent',
        domain=[
            ('uuid', '=', Eval('uuid')),
            ('parent', '=', None),
            ('calendar', '=', Eval('calendar')),
        ],
        ondelete='CASCADE', depends=['uuid', 'calendar'])
    recurrence = fields.DateTime('Recurrence', states={
        'invisible': ~Eval('_parent_parent'),
        'required': Bool(Eval('_parent_parent')),
    }, depends=['parent'])
    vevent = fields.Binary('vevent')

    @classmethod
    def __setup__(cls):
        super(Event, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('uuid_recurrence_uniq',
                Unique(t, t.uuid, t.calendar, t.recurrence),
                'UUID and recurrence must be unique in a calendar.'),
        ]

    @staticmethod
    def default_uuid():
        return str(uuid.uuid4())

    @staticmethod
    def default_sequence():
        return 0

    @staticmethod
    def default_classification():
        return 'public'

    @staticmethod
    def default_transp():
        return 'opaque'

    @staticmethod
    def timezones():
        return [(x, x) for x in pytz.common_timezones] + [('', '')]

    @property
    def _fbtype(self):
        '''
        Return the freebusy type for give transparent and status
        '''
        if self.transp == 'opaque':
            if not self.status or self.status == 'confirmed':
                fbtype = 'BUSY'
            elif self.status == 'cancelled':
                fbtype = 'FREE'
            elif self.status == 'tentative':
                fbtype = 'BUSY-TENTATIVE'
            else:
                fbtype = 'BUSY'
        else:
            fbtype = 'FREE'
        return fbtype

    @classmethod
    def validate(cls, events):
        super(Event, cls).validate(events)
        for event in events:
            event.check_recurrence()

    def check_recurrence(self):
        '''
        Check the recurrence is not recurrent.
        '''
        if self.parent:
            if self.rdates \
                    or self.rrules \
                    or self.exdates \
                    or self.exrules \
                    or self.occurences:
                raise InvalidRecurrence(
                    gettext('health_caldav.msg_invalid_recurrence'))

    @classmethod
    def view_attributes(cls):
        return [('//page[@id="occurences"]', 'states', {
            'invisible': Bool(Eval('_parent_parent')),
        })]

    @classmethod
    def create(cls, vlist):
        pool = Pool()
        Calendar = pool.get('calendar.calendar')
        Collection = pool.get('webdav.collection')

        events = super(Event, cls).create(vlist)
        for event in events:
            if (event.calendar.owner
                    and (event.organizer == event.calendar.owner.email
                         or (event.parent
                             and event.parent.organizer ==
                             event.parent.calendar.owner.email))):
                if event.organizer == event.calendar.owner.email:
                    attendee_emails = [x.email for x in event.attendees
                                       if x.status != 'declined'
                                       and x.email != event.organizer]
                else:
                    attendee_emails = [x.email for x in event.parent.attendees
                                       if x.status != 'declined'
                                       and x.email != event.parent.organizer]
                if attendee_emails:
                    with Transaction().set_user(0):
                        calendars = Calendar.search([
                            ('owner.email', 'in', attendee_emails),
                        ])
                        if not event.recurrence:
                            for calendar in calendars:
                                new_event, = cls.copy([event], default={
                                    'calendar': calendar.id,
                                    'occurences': None,
                                    'uuid': event.uuid,
                                })
                                for occurence in event.occurences:
                                    cls.copy([occurence], default={
                                        'calendar': calendar.id,
                                        'parent': new_event.id,
                                        'uuid': occurence.uuid,
                                    })
                        else:
                            parents = cls.search([
                                ('uuid', '=', event.uuid),
                                ('calendar.owner.email', 'in',
                                 attendee_emails),
                                ('id', '!=', event.id),
                                ('recurrence', '=', None),
                            ])
                            for parent in parents:
                                cls.copy([event], default={
                                    'calendar': parent.calendar.id,
                                    'parent': parent.id,
                                    'uuid': event.uuid,
                                })
        # Restart the cache for event
        Collection._event_cache.clear()
        return events

    def _event2update(self):
        return {
            'summary': self.summary,
            'description': self.description,
            'all_day': self.all_day,
            'dtstart': self.dtstart,
            'dtend': self.dtend,
            'location': self.location.id if self.location else None,
            'status': self.status,
            'organizer': self.organizer,
            'rdates': [('delete', [r.id for r in self.rdates])]
            + [('create', [rdate._date2update()
                           for rdate in self.rdates])],
            'exdates': [('delete', [r.id for r in self.exdates])]
            + [('create', [exdate._date2update()
                           for exdate in self.exdates])],
            'rrules': [('delete', [r.id for r in self.rrules])]
            + [('create', [rrule._date2update()
                           for rrule in self.rrules])],
            'exrules': [('delete', [r.id for r in self.exrules])]
            + [('create', [exrule._date2update()
                           for exrule in self.exrules])],
        }

    @classmethod
    def write(cls, *args):
        pool = Pool()
        Calendar = pool.get('calendar.calendar')
        Collection = pool.get('webdav.collection')
        transaction = Transaction()
        cursor = transaction.connection.cursor()

        actions = iter(args)
        args = []
        for events, values in zip(actions, actions):
            values = values.copy()
            if 'sequence' in values:
                del values['sequence']
            args.extend((events, values))

        super(Event, cls).write(*args)

        table = cls.__table__()
        for sub_ids in grouped_slice(events, transaction.database.IN_MAX):
            red_sql = reduce_ids(table.id, sub_ids)
            cursor.execute(*table.update(
                columns=[table.sequence],
                values=[table.sequence + 1],
                where=red_sql))

        actions = iter(args)
        for events, values in zip(actions, actions):
            if not values:
                continue
            for event in events:
                if (event.calendar.owner
                        and (event.organizer == event.calendar.owner.email
                             or (event.parent
                                 and event.parent.organizer
                                 == event.calendar.owner.email))):
                    if event.organizer == event.calendar.owner.email:
                        attendee_emails = [x.email for x in event.attendees
                                           if x.status != 'declined'
                                           and x.email != event.organizer]
                    else:
                        attendee_emails = [x.email
                                           for x in event.parent.attendees
                                           if x.status != 'declined'
                                           and x.email !=
                                           event.parent.organizer]
                    with Transaction().set_user(0):
                        events2 = cls.search([
                            ('uuid', '=', event.uuid),
                            ('id', '!=', event.id),
                            ('recurrence', '=', event.recurrence),
                        ])
                        for event2 in events2[:]:
                            if event2.calendar.owner.email in attendee_emails:
                                attendee_emails.remove(
                                    event2.calendar.owner.email)
                            else:
                                events2.remove(event2)
                                cls.delete([event2])
                        if events2:
                            cls.write(events2, event._event2update())
                    if attendee_emails:
                        with Transaction().set_user(0):
                            calendars = Calendar.search([
                                ('owner.email', 'in', attendee_emails),
                            ])
                            if not event.recurrence:
                                for calendar in calendars:
                                    new_event, = cls.copy([event], default={
                                        'calendar': calendar.id,
                                        'occurences': None,
                                        'uuid': event.uuid,
                                    })
                                    for occurence in event.occurences:
                                        cls.copy([occurence], default={
                                            'calendar': calendar.id,
                                            'parent': new_event.id,
                                            'uuid': occurence.uuid,
                                        })
                            else:
                                parents = cls.search([
                                    ('uuid', '=', event.uuid),
                                    ('calendar.owner.email', 'in',
                                     attendee_emails),
                                    ('id', '!=', event.id),
                                    ('recurrence', '=', None),
                                ])
                                for parent in parents:
                                    cls.copy([event], default={
                                        'calendar': parent.calendar.id,
                                        'parent': parent.id,
                                        'uuid': event.uuid,
                                    })
        # Restart the cache for event
        Collection._event_cache.clear()

    @classmethod
    def copy(cls, events, default=None):
        if default is None:
            default = {}

        new_events = []
        for event in events:
            current_default = default.copy()
            current_default.setdefault('uuid', cls.default_uuid())
            new_events.extend(super(Event, cls).copy([event],
                              default=current_default))
        return new_events

    @classmethod
    def delete(cls, events):
        pool = Pool()
        Attendee = pool.get('calendar.event.attendee')
        Collection = pool.get('webdav.collection')

        for event in events:
            if (event.calendar.owner
                    and (event.organizer == event.calendar.owner.email
                         or (event.parent
                             and event.parent.organizer
                             == event.calendar.owner.email))):
                if event.organizer == event.calendar.owner.email:
                    attendee_emails = [x.email for x in event.attendees
                                       if x.email != event.organizer]
                else:
                    attendee_emails = [x.email for x in event.parent.attendees
                                       if x.email != event.parent.organizer]
                if attendee_emails:
                    with Transaction().set_user(0):
                        cls.delete(cls.search([
                            ('uuid', '=', event.uuid),
                            ('calendar.owner.email', 'in',
                             attendee_emails),
                            ('id', '!=', event.id),
                            ('recurrence', '=', event.recurrence),
                        ]))
            elif event.organizer \
                    or (event.parent and event.parent.organizer):
                if event.organizer:
                    organizer = event.organizer
                else:
                    organizer = event.parent.organizer
                with Transaction().set_user(0):
                    events2 = cls.search([
                        ('uuid', '=', event.uuid),
                        ('calendar.owner.email', '=', organizer),
                        ('id', '!=', event.id),
                        ('recurrence', '=', event.recurrence),
                    ], limit=1)
                    if events2:
                        event2, = events2
                        for attendee in event2.attendees:
                            if attendee.email == event.calendar.owner.email:
                                Attendee.write([attendee], {
                                    'status': 'declined',
                                })
        super(Event, cls).delete(events)
        # Restart the cache for event
        Collection._event_cache.clear()

    @classmethod
    def ical2values(cls, event_id, ical, calendar_id, vevent=None):
        '''
        Convert iCalendar to values for create or write with
        the event id for write or None for create
        '''
        pool = Pool()
        Attendee = pool.get('calendar.event.attendee')
        Category = pool.get('calendar.category')
        Location = pool.get('calendar.location')
        Alarm = pool.get('calendar.event.alarm')
        Rdate = pool.get('calendar.event.rdate')
        Exdate = pool.get('calendar.event.exdate')
        Rrule = pool.get('calendar.event.rrule')
        Exrule = pool.get('calendar.event.exrule')

        vevents = []
        if not vevent:
            vevent = ical.vevent

            for i in ical.getChildren():
                if i.name == 'VEVENT' \
                        and i != vevent:
                    vevents.append(i)

        event = None
        if event_id:
            event = cls(event_id)

        res = {}
        if not event:
            if hasattr(vevent, 'uid'):
                res['uuid'] = vevent.uid.value
            else:
                res['uuid'] = str(uuid.uuid4())
        if hasattr(vevent, 'summary'):
            res['summary'] = vevent.summary.value
        else:
            res['summary'] = None
        if hasattr(vevent, 'description'):
            res['description'] = vevent.description.value
        else:
            res['description'] = None
        if not isinstance(vevent.dtstart.value, datetime.datetime):
            res['all_day'] = True
            res['dtstart'] = datetime.datetime.combine(
                vevent.dtstart.value,
                datetime.time())
        else:
            res['all_day'] = False
            if vevent.dtstart.value.tzinfo:
                res['dtstart'] = vevent.dtstart.value.astimezone(tzlocal)
            else:
                res['dtstart'] = vevent.dtstart.value
        if hasattr(vevent, 'dtend'):
            if not isinstance(vevent.dtend.value, datetime.datetime):
                res['dtend'] = datetime.datetime.combine(vevent.dtend.value,
                                                         datetime.time())
            else:
                if vevent.dtend.value.tzinfo:
                    res['dtend'] = vevent.dtend.value.astimezone(tzlocal)
                else:
                    res['dtend'] = vevent.dtend.value
        elif hasattr(vevent, 'duration') and hasattr(vevent, 'dtstart'):
            res['dtend'] = vevent.dtstart.value + vevent.duration.value
        else:
            res['dtend'] = None
        if hasattr(vevent, 'recurrence-id'):
            if not isinstance(vevent.recurrence_id.value, datetime.datetime):
                res['recurrence'] = datetime.datetime.combine(
                    vevent.recurrence_id.value, datetime.time()
                ).replace(tzinfo=tzlocal)
            else:
                if vevent.recurrence_id.value.tzinfo:
                    res['recurrence'] = \
                        vevent.recurrence_id.value.astimezone(tzlocal)
                else:
                    res['recurrence'] = vevent.recurrence_id.value
        else:
            res['recurrence'] = None
        if hasattr(vevent, 'status'):
            res['status'] = vevent.status.value.lower()
        else:
            res['status'] = ''
        if event:
            res['categories'] = [('remove', [c.id for c in event.categories])]
        if hasattr(vevent, 'categories'):
            with Transaction().set_context(active_test=False):
                categories = Category.search([
                    ('name', 'in', [x for x in vevent.categories.value]),
                ])
            category_names2ids = {}
            for category in categories:
                category_names2ids[category.name] = category.id
            to_create = []
            for category in vevent.categories.value:
                if category not in category_names2ids:
                    to_create.append({
                        'name': category,
                    })
            if to_create:
                categories += Category.create(to_create)
            res['categories'] += [('add', list(map(int, categories)))]
        if hasattr(vevent, 'class'):
            if getattr(vevent, 'class').value.lower() in \
                    dict(cls.classification.selection):
                res['classification'] = getattr(vevent, 'class').value.lower()
            else:
                res['classification'] = 'private'
        else:
            res['classification'] = 'public'
        if hasattr(vevent, 'location'):
            with Transaction().set_context(active_test=False):
                locations = Location.search([
                    ('name', '=', vevent.location.value),
                ], limit=1)
            if not locations:
                location, = Location.create([{
                    'name': vevent.location.value,
                }])
            else:
                location, = locations
            res['location'] = location.id
        else:
            res['location'] = None

        res['calendar'] = calendar_id

        if hasattr(vevent, 'transp'):
            res['transp'] = vevent.transp.value.lower()
        else:
            res['transp'] = 'opaque'

        if hasattr(vevent, 'organizer'):
            if vevent.organizer.value.lower().startswith('mailto:'):
                res['organizer'] = vevent.organizer.value[7:]
            else:
                res['organizer'] = vevent.organizer.value
        else:
            res['organizer'] = None

        attendees_todel = {}
        if event:
            for attendee in event.attendees:
                attendees_todel[attendee.email] = attendee.id
        res['attendees'] = []
        if hasattr(vevent, 'attendee'):
            to_create = []
            while vevent.attendee_list:
                attendee = vevent.attendee_list.pop()
                vals = Attendee.attendee2values(attendee)
                if vals['email'] in attendees_todel:
                    res['attendees'].append(('write',
                                            [attendees_todel[vals['email']]],
                                            vals))
                    del attendees_todel[vals['email']]
                else:
                    to_create.append(vals)
            if to_create:
                res['attendees'].append(('create', to_create))
        res['attendees'].append(('delete', list(attendees_todel.values())))

        res['rdates'] = []
        if event:
            res['rdates'].append(('delete', [x.id for x in event.rdates]))
        if hasattr(vevent, 'rdate'):
            to_create = []
            while vevent.rdate_list:
                rdate = vevent.rdate_list.pop()
                to_create += [Rdate.date2values(d) for d in rdate.value]
            if to_create:
                res['rdates'].append(('create', to_create))

        res['exdates'] = []
        if event:
            res['exdates'].append(('delete', [x.id for x in event.exdates]))
        if hasattr(vevent, 'exdate'):
            to_create = []
            while vevent.exdate_list:
                exdate = vevent.exdate_list.pop()
                to_create += [Exdate.date2values(d) for d in exdate.value]
            if to_create:
                res['exdates'].append(('create', to_create))

        res['rrules'] = []
        if event:
            res['rrules'].append(('delete', [x.id for x in event.rrules]))
        if hasattr(vevent, 'rrule'):
            to_create = []
            while vevent.rrule_list:
                rrule = vevent.rrule_list.pop()
                to_create.append(Rrule.rule2values(rrule))
            if to_create:
                res['rrules'].append(('create', to_create))

        res['exrules'] = []
        if event:
            res['exrules'].append(('delete', [x.id for x in event.exrules]))
        if hasattr(vevent, 'exrule'):
            to_create = []
            while vevent.exrule_list:
                exrule = vevent.exrule_list.pop()
                to_create.append(Exrule.rule2values(exrule))
            if to_create:
                res['exrules'].append(('create', to_create))

        if event:
            res.setdefault('alarms', [])
            res['alarms'].append(('delete', [x.id for x in event.alarms]))
        if hasattr(vevent, 'valarm'):
            res.setdefault('alarms', [])
            to_create = []
            while vevent.valarm_list:
                valarm = vevent.valarm_list.pop()
                to_create.append(Alarm.valarm2values(valarm))
            if to_create:
                res['alarms'].append(('create', to_create))

        if hasattr(ical, 'vtimezone'):
            if ical.vtimezone.tzid.value in pytz.common_timezones:
                res['timezone'] = ical.vtimezone.tzid.value
            else:
                for timezone in pytz.common_timezones:
                    if ical.vtimezone.tzid.value.endswith(timezone):
                        res['timezone'] = timezone

        res['vevent'] = vevent.serialize().encode()

        occurences_todel = []
        if event:
            occurences_todel = [x.id for x in event.occurences]
        to_create = []
        for vevent in vevents:
            event_id = None
            vals = cls.ical2values(event_id, ical, calendar_id, vevent=vevent)
            if event:
                for occurence in event.occurences:
                    if vals['recurrence'] == \
                            occurence.recurrence.replace(tzinfo=tzlocal):
                        event_id = occurence.id
                        occurences_todel.remove(occurence.id)
            if event:
                vals['uuid'] = event.uuid
            else:
                vals['uuid'] = res['uuid']
            res.setdefault('occurences', [])
            if event_id:
                res['occurences'].append(('write', [event_id], vals))
            else:
                to_create.append(vals)
        if to_create:
            res['occurences'].append(('create', to_create))
        if occurences_todel:
            res.setdefault('occurences', [])
            res['occurences'].insert(0, ('delete', occurences_todel))
        return res

    def event2ical(self):
        '''
        Return an iCalendar instance of vobject for event
        '''
        if self.timezone:
            tzevent = dateutil.tz.gettz(self.timezone)
        else:
            tzevent = tzlocal

        ical = vobject.iCalendar()
        vevent = ical.add('vevent')
        if self.vevent:
            ical.vevent = vobject.readOne(self.vevent.decode())
            vevent = ical.vevent
            ical.vevent.transformToNative()
        if self.summary:
            if not hasattr(vevent, 'summary'):
                vevent.add('summary')
            vevent.summary.value = self.summary
        elif hasattr(vevent, 'summary'):
            del vevent.summary
        if self.description:
            if not hasattr(vevent, 'description'):
                vevent.add('description')
            vevent.description.value = self.description
        elif hasattr(vevent, 'description'):
            del vevent.description
        if not hasattr(vevent, 'dtstart'):
            vevent.add('dtstart')
        if self.all_day:
            vevent.dtstart.value = self.dtstart.date()
        else:
            vevent.dtstart.value = self.dtstart.replace(tzinfo=tzlocal)\
                .astimezone(tzevent)
        if self.dtend:
            if not hasattr(vevent, 'dtend'):
                vevent.add('dtend')
            if self.all_day:
                vevent.dtend.value = self.dtend.date()
            else:
                vevent.dtend.value = self.dtend.replace(tzinfo=tzlocal)\
                    .astimezone(tzevent)
        elif hasattr(vevent, 'dtend'):
            del vevent.dtend
        if not hasattr(vevent, 'created'):
            vevent.add('created')
        vevent.created.value = self.create_date.replace(tzinfo=tzlocal)
        if not hasattr(vevent, 'dtstamp'):
            vevent.add('dtstamp')
        date = self.write_date or self.create_date
        vevent.dtstamp.value = date.replace(tzinfo=tzlocal)
        if not hasattr(vevent, 'last-modified'):
            vevent.add('last-modified')
        vevent.last_modified.value = date.replace(tzinfo=tzlocal)
        if self.recurrence and self.parent:
            if not hasattr(vevent, 'recurrence-id'):
                vevent.add('recurrence-id')
            if self.all_day:
                vevent.recurrence_id.value = self.recurrence.date()
            else:
                vevent.recurrence_id.value = self.recurrence\
                    .replace(tzinfo=tzlocal).astimezone(tzevent)
        elif hasattr(vevent, 'recurrence-id'):
            del vevent.recurrence_id
        if self.status:
            if not hasattr(vevent, 'status'):
                vevent.add('status')
            vevent.status.value = self.status.upper()
        elif hasattr(vevent, 'status'):
            del vevent.status
        if not hasattr(vevent, 'uid'):
            vevent.add('uid')
        vevent.uid.value = self.uuid
        if not hasattr(vevent, 'sequence'):
            vevent.add('sequence')
        vevent.sequence.value = str(self.sequence) or '0'
        if self.categories:
            if not hasattr(vevent, 'categories'):
                vevent.add('categories')
            vevent.categories.value = [x.name for x in self.categories]
        elif hasattr(vevent, 'categories'):
            del vevent.categories
        if not hasattr(vevent, 'class'):
            vevent.add('class')
            getattr(vevent, 'class').value = self.classification.upper()
        elif getattr(vevent, 'class').value.lower() in \
                dict(self.__class__.classification.selection):
            getattr(vevent, 'class').value = self.classification.upper()
        if self.location:
            if not hasattr(vevent, 'location'):
                vevent.add('location')
            vevent.location.value = self.location.name
        elif hasattr(vevent, 'location'):
            del vevent.location

        if not hasattr(vevent, 'transp'):
            vevent.add('transp')
        vevent.transp.value = self.transp.upper()

        if self.organizer:
            if not hasattr(vevent, 'organizer'):
                vevent.add('organizer')
            vevent.organizer.value = 'MAILTO:' + self.organizer
        elif hasattr(vevent, 'organizer'):
            del vevent.organizer

        vevent.attendee_list = []
        for attendee in self.attendees:
            vevent.attendee_list.append(attendee.attendee2attendee())

        if self.rdates:
            vevent.add('rdate')
            vevent.rdate.value = []
            for rdate in self.rdates:
                vevent.rdate.value.append(rdate.date2date())

        if self.exdates:
            vevent.add('exdate')
            vevent.exdate.value = []
            for exdate in self.exdates:
                vevent.exdate.value.append(exdate.date2date())

        if self.rrules:
            for rrule in self.rrules:
                vevent.add('rrule').value = rrule.rule2rule()

        if self.exrules:
            for exrule in self.exrules:
                vevent.add('exrule').value = exrule.rule2rule()

        vevent.valarm_list = []
        for alarm in self.alarms:
            valarm = alarm.alarm2valarm()
            if valarm:
                vevent.valarm_list.append(valarm)

        for occurence in self.occurences:
            oical = occurence.event2ical()
            ical.vevent_list.append(oical.vevent)
        return ical


class EventCategory(ModelSQL):
    'Event - Category'
    __name__ = 'calendar.event-calendar.category'
    event = fields.Many2One(
        'calendar.event', 'Event', ondelete='CASCADE',
        required=True)
    category = fields.Many2One(
        'calendar.category', 'Category',
        ondelete='CASCADE', required=True)


class AlarmMixin:
    __slots__ = ()
    valarm = fields.Binary('valarm')

    @classmethod
    def valarm2values(cls, valarm):
        '''
        Convert a valarm object into values for create or write
        '''
        return {
            'valarm': valarm.serialize(),
        }

    def alarm2valarm(self):
        '''
        Return a valarm instance of vobject for alarm
        '''
        if self.valarm:
            return vobject.readOne(str(self.valarm))


class EventAlarm(AlarmMixin, ModelSQL, ModelView):
    'Alarm'
    __name__ = 'calendar.event.alarm'
    event = fields.Many2One(
        'calendar.event', 'Event', ondelete='CASCADE',
        required=True)

    @classmethod
    def create(cls, vlist):
        Event = Pool().get('calendar.event')
        to_write = []
        for values in vlist:
            if values.get('event'):
                # Update write_date of event
                to_write.append(values['event'])
        if to_write:
            Event.write(Event.browse(to_write), {})
        return super(EventAlarm, cls).create(vlist)

    @classmethod
    def write(cls, *args):
        Event = Pool().get('calendar.event')

        actions = iter(args)
        events = []
        for event_alarms, values in zip(actions, actions):
            events += [x.event for x in event_alarms]
            if values.get('event'):
                events.append(Event(values['event']))
        if events:
            # Update write_date of event
            Event.write(events, {})
        super(EventAlarm, cls).write(*args)

    @classmethod
    def delete(cls, event_alarms):
        pool = Pool()
        Event = pool.get('calendar.event')
        events = [x.event for x in event_alarms]
        if events:
            # Update write_date of event
            Event.write(events, {})
        super(EventAlarm, cls).delete(event_alarms)


class AttendeeMixin:
    __slots__ = ()

    email = fields.Char('Email', required=True, states={
        'readonly': Eval('id', 0) > 0,
    }, depends=['id'])
    status = fields.Selection([
        ('', ''),
        ('needs-action', 'Needs Action'),
        ('accepted', 'Accepted'),
        ('declined', 'Declined'),
        ('tentative', 'Tentative'),
        ('delegated', 'Delegated'),
    ], 'Participation Status')
    attendee = fields.Binary('attendee')

    @staticmethod
    def default_status():
        return ''

    def _attendee2update(self):
        return {
            'status': self.status,
        }

    @classmethod
    def attendee2values(cls, attendee):
        '''
        Convert a attendee object into values for create or write
        '''
        res = {}
        if attendee.value.lower().startswith('mailto:'):
            res['email'] = attendee.value[7:]
        else:
            res['email'] = attendee.value
        res['status'] = ''
        if hasattr(attendee, 'partstat_param'):
            if attendee.partstat_param.lower() in dict(cls.status.selection):
                res['status'] = attendee.partstat_param.lower()
        res['attendee'] = attendee.serialize().encode('utf-8')
        return res

    def attendee2attendee(self):
        '''
        Return a attendee instance of vobject for attendee
        '''
        res = None
        if self.attendee:
            res = vobject.base.textLineToContentLine(
                self.attendee.decode('utf-8').replace('\r\n ', ''))
        else:
            res = vobject.base.ContentLine('ATTENDEE', [], '')

        selection = dict(self.__class__.status.selection)
        if self.status:
            if hasattr(res, 'partstat_param'):
                if res.partstat_param.lower() in selection:
                    res.partstat_param = self.status.upper()
            else:
                res.partstat_param = self.status.upper()
        elif hasattr(res, 'partstat_param'):
            if res.partstat_param.lower() in selection:
                del res.partstat_param

        res.value = 'MAILTO:' + self.email
        return res


class EventAttendee(AttendeeMixin, ModelSQL, ModelView):
    'Attendee'
    __name__ = 'calendar.event.attendee'
    event = fields.Many2One(
        'calendar.event', 'Event', ondelete='CASCADE',
        required=True)

    @classmethod
    def create(cls, vlist):
        Event = Pool().get('calendar.event')
        to_write = []
        for values in vlist:
            if values.get('event'):
                # Update write_date of event
                to_write.append(values['event'])

        if to_write:
            Event.write(Event.browse(to_write), {})
        event_attendees = super(EventAttendee, cls).create(vlist)
        for event_attendee in event_attendees:
            event = event_attendee.event
            if (event.calendar.owner
                    and (event.organizer == event.calendar.owner.email
                         or (event.parent
                             and event.parent.organizer ==
                             event.parent.calendar.owner.email))):
                if event.organizer == event.calendar.owner.email:
                    attendee_emails = [x.email for x in event.attendees
                                       if x.email != event.organizer]
                else:
                    attendee_emails = [x.email for x in event.parent.attendees
                                       if x.email != event.parent.organizer]
                if attendee_emails:
                    with Transaction().set_user(0):
                        events = Event.search([
                            ('uuid', '=', event.uuid),
                            ('calendar.owner.email', 'in',
                             attendee_emails),
                            ('id', '!=', event.id),
                            ('recurrence', '=', event.recurrence),
                        ])
                        for event in events:
                            cls.copy([event_attendee], default={
                                'event': event.id,
                            })
        return event_attendees

    @classmethod
    def write(cls, *args):
        Event = Pool().get('calendar.event')

        actions = iter(args)
        args = []
        events = []
        for event_attendees, values in zip(actions, actions):
            events += [x.event for x in event_attendees]
            if values.get('event'):
                events.append(Event(values['event']))
            if 'email' in values:
                values = values.copy()
                del values['email']
            args.extend((event_attendees, values))

        if events:
            # Update write_date of event
            Event.write(events, {})

        super(EventAttendee, cls).write(*args)

        for event_attendee in sum(args[::2], []):
            event = event_attendee.event
            if (event.calendar.owner
                    and (event.organizer == event.calendar.owner.email
                         or (event.parent
                             and event.parent.organizer
                             == event.calendar.owner.email))):
                if event.organizer == event.calendar.owner.email:
                    attendee_emails = [x.email for x in event.attendees
                                       if x.email != event.organizer]
                else:
                    attendee_emails = [x.email for x in event.parent.attendees
                                       if x.email != event.parent.organizer]
                if attendee_emails:
                    with Transaction().set_user(0):
                        other_attendees = cls.search([
                            ('event.uuid', '=', event.uuid),
                            ('event.calendar.owner.email', 'in',
                             attendee_emails),
                            ('id', '!=', event_attendee.id),
                            ('event.recurrence', '=',
                             event.recurrence),
                            ('email', '=', event_attendee.email),
                        ])
                        cls.write(other_attendees,
                                  event_attendee._attendee2update())

    @classmethod
    def delete(cls, event_attendees):
        pool = Pool()
        Event = pool.get('calendar.event')

        events = [x.event for x in event_attendees]
        if events:
            # Update write_date of event
            Event.write(events, {})

        for attendee in event_attendees:
            event = attendee.event
            if (event.calendar.owner
                    and (event.organizer == event.calendar.owner.email
                         or (event.parent
                             and event.parent.organizer
                             == event.calendar.owner.email))):
                if event.organizer == event.calendar.owner.email:
                    attendee_emails = [x.email for x in event.attendees
                                       if x.email != event.organizer]
                else:
                    attendee_emails = [x.email for x in event.parent.attendees
                                       if x.email != event.parent.organizer]
                if attendee_emails:
                    with Transaction().set_user(0):
                        attendees = cls.search([
                            ('event.uuid', '=', event.uuid),
                            ('event.calendar.owner.email', 'in',
                             attendee_emails),
                            ('id', '!=', attendee.id),
                            ('event.recurrence', '=',
                             event.recurrence),
                            ('email', '=', attendee.email),
                        ])
                        cls.delete(attendees)
            elif (event.calendar.owner
                    and ((event.organizer
                          or (event.parent and event.parent.organizer))
                         and attendee.email == event.calendar.owner.email)):
                if event.organizer:
                    organizer = event.organizer
                else:
                    organizer = event.parent.organizer
                with Transaction().set_user(0):
                    attendees = cls.search([
                        ('event.uuid', '=', event.uuid),
                        ('event.calendar.owner.email', '=', organizer),
                        ('id', '!=', attendee.id),
                        ('event.recurrence', '=', event.recurrence),
                        ('email', '=', attendee.email),
                    ])
                    if attendees:
                        cls.write(attendees, {
                            'status': 'declined',
                        })
        super(EventAttendee, cls).delete(event_attendees)


class DateMixin:
    __slots__ = ()
    # _rec_name = 'datetime'
    # _rec_name needs to be char or  txt

    def get_rec_name(self, name):
        if self.datetime:
            return str(self.datetime)

    date = fields.Boolean(
        'Is Date',
        help='Ignore time of field "Date", but handle as date only.')
    datetime = fields.DateTime('Date', required=True)

    def _date2update(self):
        return {
            'date': self.date,
            'datetime': self.datetime,
        }

    @staticmethod
    def date2values(date):
        '''
        Convert a date object into values for create or write
        '''
        res = {}
        if not isinstance(date, datetime.datetime):
            res['date'] = True
            res['datetime'] = datetime.datetime.combine(date,
                                                        datetime.time())
        else:
            res['date'] = False
            if date.tzinfo:
                res['datetime'] = date.astimezone(tzlocal)
            else:
                res['datetime'] = date
        return res

    def date2date(self):
        '''
        Return a datetime for date
        '''
        if self.date:
            return self.date.datetime.date()
        else:
            # Convert to UTC as sunbird doesn't handle tzid
            return self.datetime.replace(tzinfo=tzlocal).astimezone(tzutc)


class EventRDate(DateMixin, ModelSQL, ModelView):
    'Recurrence Date'
    __name__ = 'calendar.event.rdate'
    # _rec_name = 'datetime'

    def get_rec_name(self, name):
        if self.datetime:
            return str(self.datetime)

    event = fields.Many2One(
        'calendar.event', 'Event', ondelete='CASCADE',
        required=True)

    @classmethod
    def create(cls, vlist):
        Event = Pool().get('calendar.event')
        to_write = []
        for values in vlist:
            if values.get('event'):
                # Update write_date of event
                to_write.append(values['event'])
        if to_write:
            Event.write(Event.browse(to_write), {})
        return super(EventRDate, cls).create(vlist)

    @classmethod
    def write(cls, *args):
        Event = Pool().get('calendar.event')

        actions = iter(args)
        events = []
        for event_rdates, values in zip(actions, actions):
            events += [x.event for x in event_rdates]
            if values.get('event'):
                events.append(Event(values['event']))
        if events:
            # Update write_date of event
            Event.write(events, {})
        super(EventRDate, cls).write(*args)

    @classmethod
    def delete(cls, event_rdates):
        pool = Pool()
        Event = pool.get('calendar.event')
        events = [x.event for x in event_rdates]
        if events:
            # Update write_date of event
            Event.write(events, {})
        super(EventRDate, cls).delete(event_rdates)


class EventExDate(EventRDate):
    'Exception Date'
    __name__ = 'calendar.event.exdate'
    _table = 'calendar_event_exdate'  # Needed to override EventRDate._table


class RRuleMixin(Model):
    # _rec_name = 'freq'
    __slots__ = ()

    def get_rec_name(self, name):
        if self.freq:
            return str(self.freq)

    freq = fields.Selection([
        ('secondly', 'Secondly'),
        ('minutely', 'Minutely'),
        ('hourly', 'Hourly'),
        ('daily', 'Daily'),
        ('weekly', 'Weekly'),
        ('monthly', 'Monthly'),
        ('yearly', 'Yearly'),
    ], 'Frequency', required=True)
    until_date = fields.Boolean(
        'Is Date',
        help='Ignore time of field "Until Date", but handle as date only.')
    until = fields.DateTime('Until Date')
    count = fields.Integer('Count')
    interval = fields.Integer('Interval')
    bysecond = fields.Char('By Second')
    byminute = fields.Char('By Minute')
    byhour = fields.Char('By Hour')
    byday = fields.Char('By Day')
    bymonthday = fields.Char('By Month Day')
    byyearday = fields.Char('By Year Day')
    byweekno = fields.Char('By Week Number')
    bymonth = fields.Char('By Month')
    bysetpos = fields.Char('By Position')
    wkst = fields.Selection([
        (None, ''),
        ('su', 'Sunday'),
        ('mo', 'Monday'),
        ('tu', 'Tuesday'),
        ('we', 'Wednesday'),
        ('th', 'Thursday'),
        ('fr', 'Friday'),
        ('sa', 'Saturday'),
    ], 'Week Day', sort=False)

    @classmethod
    def __setup__(cls):
        super(RRuleMixin, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints += [
            ('until_count_only_one',
                Check(t,
                      (t.until == Null) | (t.count == Null) | (t.count == 0)),
                'Only one of "until" and "count" can be set.'),
        ]

    @classmethod
    def validate(cls, rules):
        super(RRuleMixin, cls).validate(rules)
        for rule in rules:
            rule.check_bysecond()
            rule.check_byminute()
            rule.check_byhour()
            rule.check_byday()
            rule.check_bymonthday()
            rule.check_byyearday()
            rule.check_byweekno()
            rule.check_bymonth()
            rule.check_bysetpos()

    def check_bysecond(self):
        if self.bysecond:
            for second in self.bysecond.split(','):
                try:
                    second = int(second)
                except Exception:
                    second = -1
                if not (second >= 0 and second <= 59):
                    raise InvalidBySecond(
                        gettext('health_caldav.msg_invalid_bysecond'))

    def check_byminute(self):
        if self.byminute:
            for minute in self.byminute.split(','):
                try:
                    minute = int(minute)
                except Exception:
                    minute = -1
                if not (minute >= 0 and minute <= 59):
                    raise InvalidByMinute(
                        gettext('health_caldav.msg_invalid_byminute'))

    def check_byhour(self):
        if self.byhour:
            for hour in self.byhour.split(','):
                try:
                    hour = int(hour)
                except Exception:
                    hour = -1
                if not (hour >= 0 and hour <= 23):
                    raise InvalidByHour(
                        gettext('health_caldav.msg_invalid_byhour'))

    def check_byday(self):
        if self.byday:
            for weekdaynum in self.byday.split(','):
                weekday = weekdaynum[-2:]
                if weekday not in ('SU', 'MO', 'TU', 'WE', 'TH', 'FR', 'SA'):
                    raise InvalidByDay(
                        gettext('health_caldav.msg_invalid_byday'))
                ordwk = weekday[:-2]
                if not ordwk:
                    continue
                try:
                    ordwk = int(ordwk)
                except Exception:
                    ordwk = -1
                if not (abs(ordwk) >= 1 and abs(ordwk) <= 53):
                    raise InvalidByDay(
                        gettext('health_caldav.msg_invalid_byday'))

    def check_bymonthday(self):
        if self.bymonthday:
            for monthdaynum in self.bymonthday.split(','):
                try:
                    monthdaynum = int(monthdaynum)
                except Exception:
                    monthdaynum = -100
                if not (abs(monthdaynum) >= 1 and abs(monthdaynum) <= 31):
                    raise InvalidByMonthDay(
                        gettext('health_caldav.msg_invalid_bymonthday'))

    def check_byyearday(self):
        if self.byyearday:
            for yeardaynum in self.byyearday.split(','):
                try:
                    yeardaynum = int(yeardaynum)
                except Exception:
                    yeardaynum = -1000
                if not (abs(yeardaynum) >= 1 and abs(yeardaynum) <= 366):
                    raise InvalidByYearDay(
                        gettext('health_caldav.msg_invalid_byyearday'))

    def check_byweekno(self):
        if self.byweekno:
            for weeknum in self.byweekno.split(','):
                try:
                    weeknum = int(weeknum)
                except Exception:
                    weeknum = -100
                if not (abs(weeknum) >= 1 and abs(weeknum) <= 53):
                    raise InvalidByWeekNumber(
                        gettext('health_caldav.msg_invalid_byweeknumber'))

    def check_bymonth(self):
        if self.bymonth:
            for monthnum in self.bymonth.split(','):
                try:
                    monthnum = int(monthnum)
                except Exception:
                    monthnum = -1
                if not (monthnum >= 1 and monthnum <= 12):
                    raise InvalidByMonth(
                        gettext('health_caldav.msg_invalid_bymonth'))

    def check_bysetpos(self):
        if self.bysetpos:
            for setposday in self.bysetpos.split(','):
                try:
                    setposday = int(setposday)
                except Exception:
                    setposday = -1000
                if not (abs(setposday) >= 1 and abs(setposday) <= 366):
                    raise InvalidBySetPosition(
                        gettext('health_caldav.msg_invalid_bysetpos'))

    def _rule2update(self):
        res = {}
        for field in (
            'freq', 'until_date', 'until', 'count', 'interval',
            'bysecond', 'byminute', 'byhour', 'byday', 'bymonthday',
                'byyearday', 'byweekno', 'bymonth', 'bysetpos', 'wkst'):
            res[field] = getattr(self, field)
        return res

    @classmethod
    def rule2values(cls, rule):
        '''
        Convert a rule object into values for create or write
        '''
        res = {}
        for attr in str(rule.value).replace('\\', '').split(';'):
            field, value = attr.split('=')
            field = field.lower()
            if field == 'until':
                try:
                    value = vobject.icalendar.stringToDateTime(value)
                except Exception:
                    value = vobject.icalendar.stringToDate(value)
                if not isinstance(value, datetime.datetime):
                    res['until_date'] = True
                    res['until'] = datetime.datetime.combine(value,
                                                             datetime.time())
                else:
                    res['until_date'] = False
                    if value.tzinfo:
                        res['until'] = value.astimezone(tzlocal)
                    else:
                        res['until'] = value
            elif field in ('freq', 'wkst'):
                res[field] = value.lower()
            else:
                res[field] = value
        return res

    def rule2rule(self):
        '''
        Return a rule string for rule
        '''
        res = 'FREQ=' + self.freq.upper()
        if self.until:
            res += ';UNTIL='
            if self.until_date:
                res += vobject.icalendar.dateToString(self.until.date())
            else:
                res += vobject.icalendar.dateTimeToString(
                    self.until
                    .replace(tzinfo=tzlocal).astimezone(tzutc),
                    convertToUTC=True)
        elif self.count:
            res += ';COUNT=' + str(self.count)
        for field in ('freq', 'wkst'):
            if getattr(self, field):
                res += ';' + field.upper() + '=' + getattr(self, field).upper()
        for field in (
            'interval', 'bysecond', 'byminute', 'byhour',
            'byday', 'bymonthday', 'byyearday', 'byweekno',
                'bymonth', 'bysetpos'):
            if getattr(self, field):
                res += ';' + field.upper() + '=' + str(getattr(self, field))
        return res


class EventRRule(RRuleMixin, ModelSQL, ModelView):
    'Recurrence Rule'
    __name__ = 'calendar.event.rrule'
    event = fields.Many2One(
        'calendar.event', 'Event', ondelete='CASCADE',
        required=True)

    @classmethod
    def create(cls, vlist):
        Event = Pool().get('calendar.event')
        to_write = []
        for values in vlist:
            if values.get('event'):
                # Update write_date of event
                to_write.append(values['event'])
        if to_write:
            Event.write(Event.browse(to_write), {})
        return super(EventRRule, cls).create(vlist)

    @classmethod
    def write(cls, *args):
        Event = Pool().get('calendar.event')

        actions = iter(args)
        events = []
        for event_rrules, values in zip(actions, actions):
            events += [x.event for x in event_rrules]
            if values.get('event'):
                events.append(Event(values['event']))
        if events:
            # Update write_date of event
            Event.write(events, {})
        super(EventRRule, cls).write(*args)

    @classmethod
    def delete(cls, event_rrules):
        pool = Pool()
        Event = pool.get('calendar.event')
        events = [x.event for x in event_rrules]
        if events:
            # Update write_date of event
            Event.write(events, {})
        super(EventRRule, cls).delete(event_rrules)


class EventExRule(EventRRule):
    'Exception Rule'
    __name__ = 'calendar.event.exrule'
    _table = 'calendar_event_exrule'  # Needed to override EventRRule._table



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH HISTORY package                          #
#              __init__.py: Package declaration file                    #
#########################################################################

from trytond.pool import Pool
from . import report


def register():
    Pool.register(
        report.patient_evaluation_report.PatientEvaluationReport,
        module='health_history', type_='report')



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2014 Sebastian Marro <smarro@thymbra.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH CALENDAR PACKAGE                         #
#               wizard_health_calendar: main wizard file                #
#########################################################################

from datetime import timedelta, datetime, time
import pytz
from trytond.model import ModelView, fields
from trytond.wizard import Wizard, StateView, StateAction, StateTransition, \
    Button
from trytond.pyson import PYSONEncoder
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.i18n import gettext
from trytond.modules.health.core import get_institution

__all__ = ['CreateAppointmentStart', 'CreateAppointment']

from ..exceptions import (
    NoCompanyTimezone,
    EndDateBeforeStart,
    PeriodTooLong
)


class CreateAppointmentStart(ModelView):
    'Create Appointments Start'
    __name__ = 'gnuhealth.calendar.create.appointment.start'

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof',
        required=True)
    specialty = fields.Many2One(
        'gnuhealth.specialty', 'Specialty',
        required=True)
    institution = fields.Many2One(
        'gnuhealth.institution', 'Institution',
        required=True)
    date_start = fields.Date('Start Date', required=True)
    date_end = fields.Date('End Date', required=True)
    time_start = fields.Time('Start Time', required=True, format='%H:%M')
    time_end = fields.Time('End Time', required=True, format='%H:%M')
    appointment_minutes = fields.Integer('Appointment Minutes', required=True)
    monday = fields.Boolean('Monday')
    tuesday = fields.Boolean('Tuesday')
    wednesday = fields.Boolean('Wednesday')
    thursday = fields.Boolean('Thursday')
    friday = fields.Boolean('Friday')
    saturday = fields.Boolean('Saturday')
    sunday = fields.Boolean('Sunday')

    @staticmethod
    def default_institution():
        return get_institution()

    @fields.depends('healthprof')
    def on_change_with_specialty(self):
        # Return the Current / Main speciality of the Health Professional
        # if this speciality has been specified in the HP record.
        if (self.healthprof and self.healthprof.main_specialty):
            specialty = self.healthprof.main_specialty.specialty.id
            return specialty


class CreateAppointment(Wizard):
    'Create Appointment'
    __name__ = 'gnuhealth.calendar.create.appointment'

    start = StateView(
        'gnuhealth.calendar.create.appointment.start',
        'health_calendar.create_appointment_start_view_form', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Create', 'create_', 'tryton-ok', default=True),
        ])
    create_ = StateTransition()
    open_ = StateAction('health.action_gnuhealth_appointment_view')

    def transition_create_(self):
        pool = Pool()
        Appointment = pool.get('gnuhealth.appointment')
        Company = pool.get('company.company')

        timezone = None
        company_id = Transaction().context.get('company')
        if company_id:
            company = Company(company_id)
            if company.timezone:
                timezone = pytz.timezone(company.timezone)
            else:
                raise NoCompanyTimezone(
                    gettext('health_calendar.no_company_timezone')
                )

        appointments = []

        # Iterate over days
        day_count = (self.start.date_end - self.start.date_start).days + 1

        # Validate dates
        if (self.start.date_start and self.start.date_end):
            if (self.start.date_end < self.start.date_start):
                raise EndDateBeforeStart(
                    gettext('health_calendar.msg_end_before_start')
                )

            if (day_count > 31):
                raise PeriodTooLong(
                    gettext('health_calendar.msg_period_too_long')
                )

        for single_date in (
            self.start.date_start + timedelta(n)
                for n in range(day_count)):
            if ((single_date.weekday() == 0 and self.start.monday)
                or (single_date.weekday() == 1 and self.start.tuesday)
                or (single_date.weekday() == 2 and self.start.wednesday)
                or (single_date.weekday() == 3 and self.start.thursday)
                or (single_date.weekday() == 4 and self.start.friday)
                or (single_date.weekday() == 5 and self.start.saturday)
                    or (single_date.weekday() == 6 and self.start.sunday)):
                # Iterate over time
                dt = datetime.combine(
                    single_date, self.start.time_start)
                dt = timezone.localize(dt)
                dt = dt.astimezone(pytz.utc)
                dt_end = datetime.combine(
                    single_date, self.start.time_end)
                dt_end = timezone.localize(dt_end)
                dt_end = dt_end.astimezone(pytz.utc)
                while dt < dt_end:
                    appointment = {
                        'healthprof': self.start.healthprof.id,
                        'speciality': self.start.specialty.id,
                        'institution': self.start.institution.id,
                        'appointment_date': dt,
                        'appointment_date_end': dt +
                        timedelta(minutes=self.start.appointment_minutes),
                        'state': 'free',
                    }
                    appointments.append(appointment)
                    dt += timedelta(minutes=self.start.appointment_minutes)
        if appointments:
            Appointment.create(appointments)
        return 'open_'

    def do_open_(self, action):
        action['pyson_domain'] = [
            ('healthprof', '=', self.start.healthprof.id),
            ('appointment_date', '>=',
                datetime.combine(self.start.date_start, time())),
            ('appointment_date', '<=',
                datetime.combine(self.start.date_end, time())),
        ]
        action['pyson_domain'] = PYSONEncoder().encode(action['pyson_domain'])
        action['name'] += ' - %s, %s' % (self.start.healthprof.name.lastname,
                                         self.start.healthprof.name.name)
        return action, {}

    def transition_open_(self):
        return 'end'



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#                  wizard_appointment_evaluation.py: wizard             #
#########################################################################

from trytond.wizard import Wizard, StateAction
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.pyson import PYSONEncoder

from trytond.i18n import gettext

from ..exceptions import NoAppointmentSelected
__all__ = ['CreateAppointmentEvaluation']


class CreateAppointmentEvaluation(Wizard):
    'Create Appointment Evaluation'
    __name__ = 'wizard.gnuhealth.appointment.evaluation'

    start_appointment_evaluation = StateAction('health.act_app_evaluation')
    start_state = 'start_appointment_evaluation'

    def do_start_appointment_evaluation(self, action):
        """ Fill in the relevant fields on the target patient evaluation
            form. The domain fields will be read-only, while the ones
            coming from the context can be manually updated
        """
        context = {}

        active_id = Transaction().context.get('active_id')

        if not active_id:
            raise NoAppointmentSelected(gettext(
                'health.msg_no_appointment_selected')
            )

        active_model = Transaction().context.get('active_model')
        action['name'] = f"{action['name']} {self.record.patient.rec_name}"

        Model = Pool().get(active_model)
        record = Model(active_id)

        action['pyson_domain'] = PYSONEncoder().encode([
            ('patient', '=', record.patient.id),
            ('appointment', '=', active_id),

        ])

        # Avoid overriding the patient evaluation default values
        # if they are not coded explicitly in the wizard.
        if record.healthprof:
            context['default_healthprof'] = record.healthprof.id

        if record.institution:
            context['default_institution'] = record.institution.id

        if record.insurance:
            context['default_insurance'] = record.insurance.id

        if record.speciality:
            context['default_specialty'] = record.speciality.id

        context.update({
            'default_urgency': record.urgency,
            'default_visit_type': record.visit_type,
            'default_evaluation_type': record.appointment_type,
        })

        action['pyson_context'] = PYSONEncoder().encode(context)
        return action, {}



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#                sequences.py: HMIS sequences for this package          #
#########################################################################

from trytond.model import (ModelView, ModelSingleton, ModelSQL,
                           ValueMixin, MultiValueMixin, fields)
# from trytond import backend
from trytond.pyson import Id
from trytond.pool import Pool
# from trytond.tools.multivalue import migrate_property

# Sequences
# The patient_sequence is no longer used
# The patient ID is taken from the person demographics
patient_sequence = fields.Many2One(
    'ir.sequence', 'Patient Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_patient'))])

patient_evaluation_sequence = fields.Many2One(
    'ir.sequence', 'Patient Evaluation Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_patient_evaluation'))])

appointment_sequence = fields.Many2One(
    'ir.sequence', 'Appointment Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_appointment'))])


prescription_sequence = fields.Many2One(
    'ir.sequence', 'Prescription Sequence', required=True,
    domain=[('sequence_type', '=', Id(
        'health', 'seq_type_gnuhealth_prescription'))])


# GNU HEALTH SEQUENCES
class GnuHealthSequences(ModelSingleton, ModelSQL, ModelView, MultiValueMixin):
    'Standard Sequences for GNU Health'
    __name__ = 'gnuhealth.sequences'

    patient_sequence = fields.MultiValue(
        patient_sequence)

    patient_evaluation_sequence = fields.MultiValue(
        patient_evaluation_sequence)

    appointment_sequence = fields.MultiValue(
        appointment_sequence)

    prescription_sequence = fields.MultiValue(
        prescription_sequence)

    @classmethod
    def default_patient_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_patient')
        except KeyError:
            return None

    @classmethod
    def default_patient_evaluation_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_patient_evaluation')
        except KeyError:
            return None

    @classmethod
    def default_appointment_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_appointment')
        except KeyError:
            return None

    @classmethod
    def default_prescription_sequence(cls, **pattern):
        pool = Pool()
        ModelData = pool.get('ir.model.data')
        try:
            return ModelData.get_id('health',
                                    'seq_gnuhealth_prescription')
        except KeyError:
            return None


class _ConfigurationValue(ModelSQL):

    _configuration_value_field = None

    @classmethod
    def __register__(cls, module_name):
        # Removed in GH 5.0
        # exist = backend.TableHandler.table_exist(cls._table)

        super(_ConfigurationValue, cls).__register__(module_name)


"""
    # Removed in GH 5.0
        if not exist:
            cls._migrate_property([], [], [])

    @classmethod
    def _migrate_property(cls, field_names, value_names, fields):
        field_names.append(cls._configuration_value_field)
        value_names.append(cls._configuration_value_field)
        migrate_property(
            'gnuhealth.sequences', field_names, cls, value_names,
            fields=fields)
"""


class PatientSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Patient Sequences setup'
    __name__ = 'gnuhealth.sequences.patient_sequence'
    patient_sequence = patient_sequence
    _configuration_value_field = 'patient_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class PatientEvaluationSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Patient Evaluation Sequence setup'
    __name__ = 'gnuhealth.sequences.patient_evaluation_sequence'
    patient_evaluation_sequence = patient_evaluation_sequence
    _configuration_value_field = 'patient_evaluation_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class AppointmentSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Appointment Sequence setup'
    __name__ = 'gnuhealth.sequences.appointment_sequence'
    appointment_sequence = appointment_sequence
    _configuration_value_field = 'appointment_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class PrescriptionSequence(_ConfigurationValue, ModelSQL, ValueMixin):
    'Prescription Sequence setup'
    __name__ = 'gnuhealth.sequences.prescription_sequence'
    prescription_sequence = prescription_sequence
    _configuration_value_field = 'prescription_sequence'

    @classmethod
    def check_xml_record(cls, records, values):
        return True



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                      HEALTH INSURANCE package                         #
#              __init__.py: Package declaration file                    #
#########################################################################

from trytond.pool import Pool
from . import health_insurance
from . import wizard


def register():
    Pool.register(
        health_insurance.InsurancePlanProductPolicy,
        health_insurance.InsurancePlan,
        health_insurance.HealthService,
        health_insurance.PatientProcedure,
        module='health_insurance', type_='model')
    Pool.register(
        wizard.wizard_health_insurance.CreateServiceInvoice,
        module='health_services', type_='wizard')



# SPDX-FileCopyrightText: 2009-2013 Bertrand Chenal
# SPDX-FileCopyrightText: 2009-2016 B2CK
# SPDX-FileCopyrightText: 2009-2016 Cédric Krier
# SPDX-FileCopyrightText: 2009-2016 Tryton Foundation <info@tryton.org>
# SPDX-FileCopyrightText: 2016-2024 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2016-2024 Luis Falcón <falcon@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import vobject
import urllib.request
import urllib.parse
import urllib.error
from urllib.parse import urlparse
from sql.functions import Extract
from sql.conditionals import Coalesce
from sql.aggregate import Max
import logging

from pywebdav.lib.errors import DAV_NotFound, DAV_Forbidden
from trytond.tools import reduce_ids, grouped_slice
from trytond.cache import Cache
from trytond.transaction import Transaction
from trytond.pool import Pool, PoolMeta

__all__ = ['Collection']

CALDAV_NS = 'urn:ietf:params:xml:ns:caldav'

logger = logging.getLogger(__name__)


def _comp_filter_domain(dtstart, dtend):
    return ['OR',
            [
                ['OR',
                 [('dtstart', '<=', dtstart),
                  ('dtend', '>=', dtstart)],
                 [('dtstart', '<=', dtend),
                  ('dtend', '>=', dtend)],
                 [('dtstart', '>=', dtstart),
                  ('dtend', '<=', dtend)],
                 [('dtstart', '>=', dtstart),
                  ('dtstart', '<=', dtend),
                  ('dtend', '=', None)]],
                ('parent', '=', None),
                ('rdates', '=', None),
                ('rrules', '=', None),
                ('exdates', '=', None),
                ('exrules', '=', None),
                ('occurences', '=', None),
            ],
            [  # TODO manage better recurring event
                ('parent', '=', None),
                ('dtstart', '<=', dtend),
                ['OR',
                 ('rdates', '!=', None),
                 ('rrules', '!=', None),
                 ('exdates', '!=', None),
                 ('exrules', '!=', None),
                 ('occurences', '!=', None),
                 ]
            ]]


class Collection(metaclass=PoolMeta):
    __name__ = "webdav.collection"
    _event_cache = Cache('webdav_collection.event')

    @staticmethod
    def calendar(uri, ics=False):
        '''
        Return the calendar id in the uri
        '''
        Calendar = Pool().get('calendar.calendar')

        if uri and uri.startswith('Calendars/'):
            calendar, uri = (uri[10:].split('/', 1) + [None])[0:2]
            if ics:
                if calendar.endswith('.ics'):
                    calendar = calendar[:-4]
                else:
                    return None
            return Calendar.get_name(calendar)

    @classmethod
    def event(cls, uri, calendar_id=False):
        '''
        Return the event id in the uri
        '''
        Event = Pool().get('calendar.event')

        key = (uri, calendar_id)
        event_id = cls._event_cache.get(key, default=-1)
        if event_id != -1:
            return event_id
        else:
            event_id = None
        if uri and uri.startswith('Calendars/'):
            calendar, event_uri = (uri[10:].split('/', 1) + [None])[0:2]
            if not calendar_id:
                calendar_id = cls.calendar(uri)
                if not calendar_id:
                    return None
            events = Event.search([
                ('calendar', '=', calendar_id),
                ('uuid', '=', event_uri[:-4]),
                ('parent', '=', None),
            ], limit=1)
            if events:
                event_id = events[0].id
        cls._event_cache.set(key, event_id)
        return event_id

    @staticmethod
    def _caldav_filter_domain_calendar(filter):
        '''
        Return a domain for caldav filter on calendar
        '''
        if not filter:
            return []
        if filter.localName == 'principal-property-search':
            return [('id', '=', 0)]
        return [('id', '=', 0)]

    @classmethod
    def _caldav_filter_domain_event(cls, filter):
        '''
        Return a domain for caldav filter on event
        '''
        res = []
        if not filter:
            return []
        if filter.localName == 'principal-property-search':
            return [('id', '=', 0)]
        elif filter.localName == 'calendar-query':
            result = []
            calendar_filter = None
            for e in filter.childNodes:
                if e.nodeType == e.TEXT_NODE:
                    continue
                if e.localName == 'filter':
                    calendar_filter = e
                    break
            if calendar_filter is None:
                return []
            for vcalendar_filter in calendar_filter.childNodes:
                if vcalendar_filter.nodeType == vcalendar_filter.TEXT_NODE:
                    continue
                if vcalendar_filter.getAttribute('name') != 'VCALENDAR':
                    return [('id', '=', 0)]
                vevent_filter = None
                for vevent_filter in vcalendar_filter.childNodes:
                    if vevent_filter.nodeType == vevent_filter.TEXT_NODE:
                        vevent_filter = None
                        continue
                    if vevent_filter.localName == 'comp-filter':
                        if vevent_filter.getAttribute('name') != 'VEVENT':
                            vevent_filter = None
                            continue
                        for comp_filter in vevent_filter.childNodes:
                            if comp_filter.localName != 'time-range':
                                continue
                            start = comp_filter.getAttribute('start')
                            if start:
                                start = vobject.icalendar.stringToDateTime(
                                    start)
                                end = comp_filter.getAttribute('end')
                            if end:
                                end = vobject.icalendar.stringToDateTime(end)

                            result.append(_comp_filter_domain(start, end))
                        break
                if vevent_filter is None:
                    return [('id', '=', 0)]
                break
            return result
        elif filter.localName == 'calendar-multiget':
            ids = []
            for e in filter.childNodes:
                if e.nodeType == e.TEXT_NODE:
                    continue
                if e.localName == 'href':
                    if not e.firstChild:
                        continue
                    uri = urlparse(e.firstChild.data).path
                    dbname, uri = (uri.lstrip('/').split('/', 1) + [None])[0:2]

                    if not dbname:
                        continue
                    dbname == urllib.parse.unquote_plus(dbname)
                    if dbname != Transaction().database.name:
                        continue
                    if uri:
                        uri = urllib.parse.unquote_plus(uri)
                    event_id = cls.event(uri)
                    if event_id:
                        ids.append(event_id)
            return [('id', 'in', ids)]
        return res

    @classmethod
    def get_childs(cls, uri, filter=None, cache=None):
        pool = Pool()
        Calendar = pool.get('calendar.calendar')
        Event = pool.get('calendar.event')

        if uri in ('Calendars', 'Calendars/'):
            domain = cls._caldav_filter_domain_calendar(filter)
            domain = [['OR',
                       ('owner', '=', Transaction().user),
                       ('read_users', '=', Transaction().user),
                       ],
                      domain]
            calendars = Calendar.search(domain)
            if cache is not None:
                cache.setdefault('_calendar', {})
                cache['_calendar'].setdefault(Calendar.__name__, {})
                for calendar in calendars:
                    cache['_calendar'][Calendar.__name__][calendar.id] = {}
            return ([x.name for x in calendars] +
                    [x.name + '.ics' for x in calendars])
        if uri and uri.startswith('Calendars/'):
            calendar_id = cls.calendar(uri)
            if calendar_id and not (uri[10:].split('/', 1) + [None])[1]:
                domain = cls._caldav_filter_domain_event(filter)
                events = Event.search([
                    ('calendar', '=', calendar_id),
                    domain,
                ])
                if cache is not None:
                    cache.setdefault('_calendar', {})
                    cache['_calendar'].setdefault(Event.__name__, {})
                    for event in events:
                        cache['_calendar'][Event.__name__][event.id] = {}
                return [x.uuid + '.ics' for x in events]
            return []
        childs = super(Collection, cls).get_childs(
            uri, filter=filter, cache=cache)
        if not uri and not filter:
            childs.append('Calendars')
        elif not uri and filter:
            if filter.localName == 'principal-property-search':
                childs.append('Calendars')
        return childs

    @classmethod
    def get_resourcetype(cls, uri, cache=None):
        from pywebdav.lib.constants import COLLECTION, OBJECT
        if uri in ('Calendars', 'Calendars/'):
            return COLLECTION
        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                return COLLECTION
            return OBJECT
        elif cls.calendar(uri, ics=True):
            return OBJECT
        return super(Collection, cls).get_resourcetype(uri, cache=cache)

    @classmethod
    def get_displayname(cls, uri, cache=None):
        Calendar = Pool().get('calendar.calendar')
        if uri in ('Calendars', 'Calendars/'):
            return 'Calendars'
        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                return Calendar(calendar_id).rec_name
            return uri.split('/')[-1]
        elif cls.calendar(uri, ics=True):
            return uri.split('/')[-1]
        return super(Collection, cls).get_displayname(uri, cache=cache)

    @classmethod
    def get_contenttype(cls, uri, cache=None):
        if cls.event(uri) \
                or cls.calendar(uri, ics=True):
            return 'text/calendar'
        return super(Collection, cls).get_contenttype(uri, cache=cache)

    @classmethod
    def get_creationdate(cls, uri, cache=None):
        Calendar = Pool().get('calendar.calendar')
        Event = Pool().get('calendar.event')
        calendar = Calendar.__table__()
        event = Event.__table__()

        calendar_id = cls.calendar(uri)
        if not calendar_id:
            calendar_id = cls.calendar(uri, ics=True)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                if cache is not None:
                    cache.setdefault('_calendar', {})
                    cache['_calendar'].setdefault(Calendar.__name__, {})
                    ids = list(cache['_calendar'][Calendar.__name__].keys())
                    if calendar_id not in ids:
                        ids.append(calendar_id)
                    elif 'creationdate' in cache['_calendar'][
                            Calendar.__name__][calendar_id]:
                        return cache['_calendar'][Calendar.__name__][
                            calendar_id]['creationdate']
                else:
                    ids = [calendar_id]
                res = None
                cursor = Transaction().connection.cursor()
                for sub_ids in grouped_slice(ids):
                    red_sql = reduce_ids(calendar.id, sub_ids)
                    cursor.execute(*calendar.select(
                        calendar.id,
                        Extract('EPOCH', calendar.create_date),
                        where=red_sql))
                    for calendar_id2, date in cursor.fetchall():
                        if calendar_id2 == calendar_id:
                            res = date
                        if cache is not None:
                            cache['_calendar'][Calendar.__name__]\
                                .setdefault(calendar_id2, {})
                            cache['_calendar'][Calendar.__name__][
                                calendar_id2]['creationdate'] = date
                if res is not None:
                    return res
            else:
                event_id = cls.event(uri, calendar_id=calendar_id)
                if event_id:
                    if cache is not None:
                        cache.setdefault('_calendar', {})
                        cache['_calendar'].setdefault(Event.__name__, {})
                        ids = list(cache['_calendar'][Event.__name__].keys())
                        if event_id not in ids:
                            ids.append(event_id)
                        elif 'creationdate' in cache['_calendar'][
                                Event.__name__][event_id]:
                            return cache['_calendar'][Event.__name__][
                                event_id]['creationdate']
                    else:
                        ids = [event_id]
                    res = None
                    cursor = Transaction().connection.cursor()
                    for sub_ids in grouped_slice(ids):
                        red_sql = reduce_ids(event.id, sub_ids)
                        cursor.execute(*event.select(
                            event.id,
                            Extract('EPOCH', event.create_date),
                            where=red_sql))
                        for event_id2, date in cursor.fetchall():
                            if event_id2 == event_id:
                                res = date
                            if cache is not None:
                                cache['_calendar'][Event.__name__]\
                                    .setdefault(event_id2, {})
                                cache['_calendar'][Event.__name__][
                                    event_id2]['creationdate'] = date
                    if res is not None:
                        return res
        return super(Collection, cls).get_creationdate(uri, cache=cache)

    @classmethod
    def get_lastmodified(cls, uri, cache=None):
        pool = Pool()
        Calendar = pool.get('calendar.calendar')
        Event = pool.get('calendar.event')
        calendar = Calendar.__table__()
        event = Event.__table__()

        transaction = Transaction()
        cursor = transaction.connection.cursor()
        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                if cache is not None:
                    cache.setdefault('_calendar', {})
                    cache['_calendar'].setdefault(Calendar.__name__, {})
                    ids = list(cache['_calendar'][Calendar.__name__].keys())
                    if calendar_id not in ids:
                        ids.append(calendar_id)
                    elif 'lastmodified' in cache['_calendar'][
                            Calendar.__name__][calendar_id]:
                        return cache['_calendar'][Calendar.__name__][
                            calendar_id]['lastmodified']
                else:
                    ids = [calendar_id]
                res = None
                for sub_ids in grouped_slice(ids):
                    red_sql = reduce_ids(calendar.id, sub_ids)
                    cursor.execute(*calendar.select(
                        calendar.id,
                        Extract('EPOCH', Coalesce(
                            calendar.write_date,
                            calendar.create_date)),
                        where=red_sql))
                    for calendar_id2, date in cursor.fetchall():
                        if calendar_id2 == calendar_id:
                            res = date
                        if cache is not None:
                            cache['_calendar'][Calendar.__name__]\
                                .setdefault(calendar_id2, {})
                            cache['_calendar'][Calendar.__name__][
                                calendar_id2]['lastmodified'] = date
                if res is not None:
                    return res
            else:
                event_id = cls.event(uri, calendar_id=calendar_id)
                if event_id:
                    if cache is not None:
                        cache.setdefault('_calendar', {})
                        cache['_calendar'].setdefault(Event.__name__, {})
                        ids = list(cache['_calendar'][Event.__name__].keys())
                        if event_id not in ids:
                            ids.append(event_id)
                        elif 'lastmodified' in cache['_calendar'][
                                Event.__name__][event_id]:
                            return cache['_calendar'][Event.__name__][
                                event_id]['lastmodified']
                    else:
                        ids = [event_id]
                    res = None
                    # for sub_ids in grouped_slice(ids,
                    #        transaction.database.IN_MAX / 2):
                    for sub_ids in grouped_slice(ids):
                        red_id_sql = reduce_ids(event.id, sub_ids)
                        red_parent_sql = reduce_ids(event.parent, sub_ids)
                        cursor.execute(*event.select(
                            Coalesce(event.parent, event.id),
                            Max(Extract('EPOCH', Coalesce(
                                event.write_date,
                                event.create_date))),
                            where=red_id_sql | red_parent_sql,
                            group_by=(event.parent, event.id)))
                        for event_id2, date in cursor.fetchall():
                            if event_id2 == event_id:
                                res = date
                            if cache is not None:
                                cache['_calendar'][Event.__name__]\
                                    .setdefault(event_id2, {})
                                cache['_calendar'][Event.__name__][
                                    event_id2]['lastmodified'] = date
                    if res is not None:
                        return res
        calendar_ics_id = cls.calendar(uri, ics=True)
        if calendar_ics_id:
            if cache is not None:
                cache.setdefault('_calendar', {})
                cache['_calendar'].setdefault(Calendar.__name__, {})
                ids = list(cache['_calendar'][Calendar.__name__].keys())
                if calendar_ics_id not in ids:
                    ids.append(calendar_ics_id)
                elif 'lastmodified ics' in cache['_calendar'][
                        Calendar.__name__][calendar_ics_id]:
                    return cache['_calendar'][Calendar.__name__][
                        calendar_ics_id]['lastmodified ics']
            else:
                ids = [calendar_ics_id]
            res = None
            for sub_ids in grouped_slice(ids):
                red_sql = reduce_ids(event.calendar, sub_ids)
                cursor.execute(*event.select(
                    event.calendar,
                    Max(Extract('EPOCH', Coalesce(
                        event.write_date,
                        event.create_date))),
                    where=red_sql,
                    group_by=event.calendar))
                for calendar_id2, date in cursor.fetchall():
                    if calendar_id2 == calendar_ics_id:
                        res = date
                    if cache is not None:
                        cache['_calendar'][Calendar.__name__]\
                            .setdefault(calendar_id2, {})
                        cache['_calendar'][Calendar.__name__][
                            calendar_id2]['lastmodified ics'] = date
            if res is not None:
                return res
        return super(Collection, cls).get_lastmodified(uri, cache=cache)

    @classmethod
    def get_data(cls, uri, cache=None):
        pool = Pool()
        Event = pool.get('calendar.event')
        Calendar = pool.get('calendar.calendar')

        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                raise DAV_NotFound
            event_id = cls.event(uri, calendar_id=calendar_id)
            if not event_id:
                raise DAV_NotFound
            ical = Event(event_id).event2ical()
            return ical.serialize()
        calendar_ics_id = cls.calendar(uri, ics=True)
        if calendar_ics_id:
            ical = Calendar(calendar_ics_id).calendar2ical()
            return ical.serialize()
        return super(Collection, cls).get_data(uri, cache=cache)

    @classmethod
    def get_calendar_description(cls, uri, cache=None):
        Calendar = Pool().get('calendar.calendar')

        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                if cache is not None:
                    cache.setdefault('_calendar', {})
                    cache['_calendar'].setdefault(Calendar.__name__, {})
                    ids = list(cache['_calendar'][Calendar.__name__].keys())
                    if calendar_id not in ids:
                        ids.append(calendar_id)
                    elif 'calendar_description' in cache['_calendar'][
                            Calendar.__name__][calendar_id]:
                        res = cache['_calendar'][Calendar.__name__][
                            calendar_id]['calendar_description']
                        if res is not None:
                            return res
                else:
                    ids = [calendar_id]
                res = None
                for calendar in Calendar.browse(ids):
                    if calendar.id == calendar_id:
                        res = calendar.description
                    if cache is not None:
                        cache['_calendar'][Calendar.__name__]\
                            .setdefault(calendar.id, {})
                        cache['_calendar'][Calendar.__name__][
                            calendar.id]['calendar_description'] = \
                            calendar.description
                if res is not None:
                    return res
        raise DAV_NotFound

    @classmethod
    def get_calendar_data(cls, uri, cache=None):
        return cls.get_data(uri, cache=cache)
        # return cls.get_data(uri, cache=cache).decode('utf-8')

    @staticmethod
    def get_calendar_home_set(uri, cache=None):
        return '/Calendars'

    @staticmethod
    def get_calendar_user_address_set(uri, cache=None):
        User = Pool().get('res.user')
        user = User(Transaction().user)
        if user.email:
            return user.email
        raise DAV_NotFound

    @staticmethod
    def get_schedule_inbox_URL(uri, cache=None):
        Calendar = Pool().get('calendar.calendar')
        user = Transaction().user

        calendars = Calendar.search([
            ('owner', '=', user),
        ], limit=1)
        if not calendars:
            # Sunbird failed with no value
            return '/Calendars'
        calendar, = calendars
        return '/Calendars/' + calendar.name

    @classmethod
    def get_schedule_outbox_URL(cls, uri, cache=None):
        return cls.get_schedule_inbox_URL(uri, cache=cache)

    @classmethod
    def get_event_hprof(cls, event):
        """ Returns the health professional ID associated to the event
        """
        Party = Pool().get('party.party')
        Hprof = Pool().get('gnuhealth.healthprofessional')

        event_owner = event.calendar.owner.id

        party, = Party.search(
            [('internal_user', '=', event_owner)], limit=1)

        hprof, = Hprof.search(
            [('name', '=', party)], limit=1)

        return hprof

    @classmethod
    def appointment_from_event(cls, values, Appointment, event):
        """ Creates the patient appointment associated to the event
            The patient ID is the first word from the summary field
        """
        Patient = Pool().get('gnuhealth.patient')
        summary = values['summary']
        desc = values['description']
        header = summary.split(" ")
        patient_id = header[0]
        header_suffix = " ".join(header[1:])
        app_date = values['dtstart']
        app_end = values['dtend']

        app = []
        app_values = {}
        comments = ''
        patient = None

        if desc:
            comments = header_suffix + '\n' + desc

        if (patient_id):
            res = Patient.search(
                [('puid', '=', patient_id)], limit=1)
            # Warning if the patient ID is not found
            if (len(res) > 0):
                patient = res[0]
            else:
                logger.error(
                    "The patient with PUID '%s' is not found.",
                    patient_id)

        # If the ID is associated to a patient
        # create the appointment related to her / him
        if (patient):

            # Get the Health professional associated to the calendar event
            hprof = cls.get_event_hprof(event)

            app_values = {
                'patient': patient,
                'healthprof': hprof,
                'event': event,
                'comments': comments,
                'state': 'confirmed',
                'appointment_date': app_date,
                'appointment_date_end': app_end
            }

            app.append(app_values)
            Appointment.create(app)

    @classmethod
    def put(cls, uri, data, content_type, cache=None):
        pool = Pool()
        Event = pool.get('calendar.event')
        Calendar = pool.get('calendar.calendar')

        Appointment = pool.get('gnuhealth.appointment')

        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                raise DAV_Forbidden
            event_id = cls.event(uri, calendar_id=calendar_id)
            if not event_id:
                # Create a new event
                # Convert to data str from binary
                if (isinstance(data, bytes)):
                    ical = vobject.readOne(data.decode())
                else:
                    ical = vobject.readOne(data)

                values = Event.ical2values(None, ical, calendar_id)
                event, = Event.create([values])

                # Create the appointment associated to the Event
                # Checks the patient name based on the first word
                # of the summary field.
                cls.appointment_from_event(values, Appointment, event)

                calendar = Calendar(calendar_id)
                return (Transaction().database.name + '/Calendars/' +
                        calendar.name + '/' + event.uuid + '.ics')
            else:
                # Update an existing event
                # Convert to data str from binary
                ical = vobject.readOne(data.decode())
                values = Event.ical2values(event_id, ical, calendar_id)
                Event.write([Event(event_id)], values)

                # Check if the event is an appointment in GNU Health
                dtstart = values['dtstart']
                dtend = values['dtend']
                description = values['description']

                # Check if there is an appointment in GNU Health
                # associated to this event
                # If it finds it, then update the current date
                # and description

                appointment = Appointment.search(
                    [('event', '=', event_id)], limit=1)
                if (appointment):
                    app = appointment[0]
                    app_vals = {'appointment_date': dtstart,
                                'appointment_date_end': dtend,
                                'comments': description}

                    Appointment.write([app], app_vals)

                return
        calendar_ics_id = cls.calendar(uri, ics=True)
        if calendar_ics_id:
            raise DAV_Forbidden
        return super(Collection, cls).put(uri, data, content_type)

    @classmethod
    def mkcol(cls, uri, cache=None):
        if uri and uri.startswith('Calendars/'):
            raise DAV_Forbidden
        return super(Collection, cls).mkcol(uri, cache=cache)

    @classmethod
    def rmcol(cls, uri, cache=None):
        Calendar = Pool().get('calendar.calendar')

        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                try:
                    Calendar.delete([Calendar(calendar_id)])
                except Exception:
                    raise DAV_Forbidden
                return 200
            raise DAV_Forbidden
        return super(Collection, cls).rmcol(uri, cache=cache)

    @classmethod
    def rm(cls, uri, cache=None):
        Event = Pool().get('calendar.event')

        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                return 403
            event_id = cls.event(uri, calendar_id=calendar_id)
            if event_id:
                try:
                    Event.delete([Event(event_id)])
                except Exception:
                    return 403
                return 200
            return 404
        calendar_ics_id = cls.calendar(uri, ics=True)
        if calendar_ics_id:
            return 403
        return super(Collection, cls).rm(uri, cache=cache)

    @classmethod
    def exists(cls, uri, cache=None):
        if uri in ('Calendars', 'Calendars/'):
            return 1
        calendar_id = cls.calendar(uri)
        if calendar_id:
            if not (uri[10:].split('/', 1) + [None])[1]:
                return 1
            if cls.event(uri, calendar_id=calendar_id):
                return 1
        calendar_ics_id = cls.calendar(uri, ics=True)
        if calendar_ics_id:
            return 1
        return super(Collection, cls).exists(uri, cache=cache)

    @classmethod
    def current_user_privilege_set(cls, uri, cache=None):
        '''
        Return the privileges of the current user for uri
        Privileges ares: create, read, write, delete
        '''
        Calendar = Pool().get('calendar.calendar')

        if uri in ('Calendars', 'Calendars/'):
            return ['create', 'read', 'write', 'delete']
        if uri and uri.startswith('Calendars/'):
            calendar_id = cls.calendar(uri)
            if calendar_id:
                calendar = Calendar(calendar_id)
                user = Transaction().user
                if user == calendar.owner.id:
                    return ['create', 'read', 'write', 'delete']
                res = []
                if user in (x.id for x in calendar.read_users):
                    res.append('read')
                if user in (x.id for x in calendar.write_users):
                    res.extend(['read', 'write', 'delete'])
                return res
            return []
        return super(Collection, cls).current_user_privilege_set(
            uri, cache=cache)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from trytond.exceptions import UserError, UserWarning
from trytond.model.exceptions import ValidationError


class NoAssociatedHealthProfessional(ValidationError):
    pass


class DischargeReasonNeeded(UserError):
    pass


class DischargeBeforeAdmission(UserError):
    pass


class BedIsNotAvailable(UserError):
    pass


class DestinationBedNotAvailable(UserError):
    pass


class NeedTimeZone(UserError):
    pass


class AdmissionMustBeToday(UserError):
    pass


class ManyRecordsChosen(UserError):
    pass


class NoRecordSelected(UserError):
    pass


class SpecialMealNeeds(UserWarning):
    pass



# SPDX-FileCopyrightText:  2024 - Wei Zhao <wei.zhao@uclouvain.be>
# SPDX-License-Identifier: GPL-3.0-or-later
#
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH IMAGING ORTHANC package                          #
#                  health_orthanc_configuration.py
#########################################################################


"""
The Configuration of Orthanc DICOM Server.
"""

from trytond.model import ModelView, ModelSQL, fields, Unique
from pyorthanc import Orthanc
from urllib.parse import urljoin
from requests.exceptions import HTTPError, RequestException
from trytond.exceptions import UserError
from trytond.pool import Pool
import logging

__all__ = ['ServerConfig']

logger = logging.getLogger(__name__)


class ServerConfig(ModelSQL, ModelView):
    """
    Orthanc Server Config

    This class is used to connect to an Orthanc DICOM server and
    to check if a connection to the corresponding domain can be established.
    """
    __name__ = "gnuhealth.orthanc.config"
    _rec_name = "label"

    label = fields.Char(
        "Label", required=True,
        readonly=True,
        help="Label for server (eg., remote)")

    domain = fields.Char(
        "URL",
        required=True,
        readonly=True,
        help="The full URL of the Orthanc server")

    proxy_domain = fields.Char(
        "Proxy URL",
        required=False,
        readonly=True,
        help="The proxy URL of the Orthanc server")

    link_base_url = fields.Function(
        fields.Char(
            "Link Base URL",
            help="Base URL for links"),
        "get_link_base_url")

    user = fields.Char(
        "Username",
        required=True,
        help="Username for Orthanc REST server")

    password = fields.Char(
        "Password",
        required=True,
        help="Password for Orthanc REST server")

    validated = fields.Boolean(
        "Validated",
        help="Whether the server details have "
        "been successfully checked")

    link = fields.Function(
        fields.Char(
            "Link",
            help="Link to Orthanc Explorer"),
        "get_link")

    last_changed_index = fields.Integer(
        "Last Changed Index",
        readonly=True,
        help="Index of last change")

    @classmethod
    def default_last_changed_index(cls):
        """
        Class method to return the default last changed index.
        """
        return -1

    def get_link_base_url(self, name):
        """
        Get the base URL for links shown in the client
        """
        url = self.proxy_domain
        if url is None or url.strip() == "":
            url = self.domain
        return "".join([url.strip().rstrip("/"), "/"])

    def get_link(self, name):
        """
        Get the full explorer link
        """
        # return urljoin(self.link_base_url, "app/explorer.html")
        return urljoin(self.link_base_url, "")

    @classmethod
    def __setup__(cls):
        """
        Set up the ServerConfig class for database access.

        This method is a class method that initializes various properties
        and constraints of the ServerConfig model. It sets up a SQL
        constraint to ensure that the ``label`` coulmn is unique.
        """
        super().__setup__()
        t = cls.__table__()
        cls._buttons.update({
            'remove_config_server': {}
        })

        cls._sql_constraints = [
            ("label_unique", Unique(t, t.label),
             "The label must be unique."),
            ("domain_unique", Unique(t, t.domain),
             "The domain must be unique."),
        ]

    @staticmethod
    def quick_check(domain, user, password):
        """
        Check if the server details are correct.

        :param domain: The domain name or IP address of the Orthanc
                       DICOM server.
        :type domain: str

        :param user: The username for authentication.
        :type user: str

        :param password: The password for authentication.
        :type password: str

        :return: ``True`` if the server details are valid,
                 ``False`` otherwise.
        :rtype: bool
        """

        try:
            client = Orthanc(
                url=domain, username=user, password=password)
            client.get_changes({'last': 1})
        except ConnectionError:
            logger.exception(
                "No connection to the server can be established."
                "Check connectivity and port."
            )
            return False
        except HTTPError as err:
            status_code = err.response.status_code
            if status_code in ServerConfig.http_error_messages:
                error_message = (
                    ServerConfig.http_error_messages[status_code] +
                    f" {domain} not reacheable"
                )
                logger.exception(error_message)
            else:
                logger.exception("Unhandled HTTP error for <%s>", domain)
            return False
        except RequestException:
            logger.exception(
                "Unhandled request error for <%s> occurred", domain)
            return False
        except BaseException:
            logger.exception(
                "Other error for <%s> occurred", domain)
            return False
        return True

    @fields.depends("domain", "user", "password")
    def on_change_with_validated(self):
        """
        Update the ``validated`` field based on the current server details.

        :return: A boolean value indicating whether the update was
                 successful or not.
        :rtype: bool

        .. note:: This method follows the Tryton Syntax. The
                  ``@fields.depends`` decorates the method to indicate
                    that this field depends on other fields. In addition,
                    ``on_change_with_`` is appended before the field name
                    to indicate that the field should change depending on
                    the parameters after ``@fields.depends``.

        .. seealso:: React to user input and Add computed fields in Tryton
                     documentation.
        """

        return self.quick_check(self.domain, self.user, self.password)

    @classmethod
    @ModelView.button
    def remove_config_server(cls, records):
        Study = Pool().get("gnuhealth.imaging_orthanc.study")

        # check which configurations we can delete
        failed_domains = []
        for config in records:
            # check if there are studies with same domain as this config
            studies = Study.search([("server", "=", config.domain)])
            if len(studies) > 0:
                failed_domains.append(config.domain)
            else:
                cls.delete([config])

        if len(failed_domains) > 0:
            raise UserError(
                ("Cannot remove the following servers "
                 "because there are studies from them: {}").format(
                     ", ".join(failed_domains)))
        return "reload"



# SPDX-FileCopyrightText: 2019-2022 Chris Zimmerman <chris@teffalump.com>
# SPDX-FileCopyrightText: 2021-2024 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2021-2024 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH IMAGING WORKLIST package                 #
#                  __init__.py: Package declaration file                #
#########################################################################

"""
Initialization module for the ``health_imaging_worklist`` module.

This module registers the necessary classes and methods of the
``health_imaging_worklist`` module and its wizard in the Tryton pool.
This allows other modules to access the functionalities provided by
the ``health_imaging_worklist`` module.
"""

from trytond.pool import Pool
from . import health_imaging_worklist


def register():
    Pool.register(
        health_imaging_worklist.WorklistTemplate,
        health_imaging_worklist.ImagingTestRequest,
        health_imaging_worklist.ImagingTest,
        health_imaging_worklist.TestResult,
        module="health_imaging_worklist",
        type_="model",
    )



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#               core.py: commonly used ojects and methods               #
#########################################################################
import pytz

from dateutil.relativedelta import relativedelta
from datetime import datetime
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.i18n import gettext

from .exceptions import (NoAssociatedHealthProfessional)

import os
import json


def get_institution_timezone():
    """
    Return timezone of current institution,
    if not found, return UTC as fallback.
    """

    Company = Pool().get('company.company')

    timezone = pytz.UTC
    company_id = Transaction().context.get('company')

    if company_id:
        company = Company(company_id)
        if company.timezone:
            timezone = pytz.timezone(company.timezone)

    return timezone


def convert_date_timezone(sdate, target=None):
    """
    Convert dates from UTC to local timezone and viceversa
    Datetime values are stored in UTC, so we need conversion
    """

    Company = Pool().get('company.company')

    institution_timezone = pytz.UTC
    company_id = Transaction().context.get('company')
    if company_id:
        company = Company(company_id)
        if company.timezone:
            institution_timezone = pytz.timezone(company.timezone)

    if (target == 'utc'):
        # Convert date to UTC timezone
        res = institution_timezone.localize(sdate).astimezone(pytz.utc)
    else:
        # Convert from UTC to institution local timezone
        res = pytz.utc.localize(sdate).astimezone(institution_timezone)
    return res


def estimated_date_from_years(years_old):
    """ returns a date of substracting the
        referred number of years from today's date
        It can be used in different context, such as to estimate
        the date of birth from a referred age in years
    """

    today = datetime.today().date()
    est_dob = today - relativedelta(years=years_old)
    return est_dob


def compute_age_from_dates(dob, deceased, dod, gender, caller, extra_date):
    """ Get the person's age.

    Calculate the current age of the patient or age at time of death.

    Returns:
    If caller == 'age': str in Y-M-D,
       caller == 'childbearing_age': boolean,
       caller == 'raw_age': [Y, M, D]

    """
    today = datetime.today().date()

    if dob:
        start = datetime.strptime(str(dob), '%Y-%m-%d')
        end = datetime.strptime(str(today), '%Y-%m-%d')

        if extra_date:
            end = datetime.strptime(str(extra_date), '%Y-%m-%d')

        if deceased and dod:
            end = datetime.strptime(
                str(dod), '%Y-%m-%d %H:%M:%S')

        rdelta = relativedelta(end, start)

        years_months_days = format_years_months_days(
            years=rdelta.years,
            months=rdelta.months,
            days=rdelta.days)
    else:
        return None

    if caller == 'age':
        return years_months_days

    elif caller == 'childbearing_age':
        if (rdelta.years >= 11
           and rdelta.years <= 55 and gender == 'f'):
            return True
        else:
            return False

    elif caller == 'raw_age':
        return [rdelta.years, rdelta.months, rdelta.days]

    else:
        return None


def format_years_months_days(years=None, months=None, days=None):
    year_str = gettext('health.msg_compute_age_from_dates_year_str')
    month_str = gettext('health.msg_compute_age_from_dates_month_str')
    day_str = gettext('health.msg_compute_age_from_dates_day_str')

    ymd_format = '{years}{sep}{year_str}{sep}' \
                 '{months}{sep}{month_str}{sep}' \
                 '{days}{sep}{day_str}{sep}'

    return ymd_format.format(
        sep='\u200b',  # Zero width space
        # Make sure output.split(sep)[0, 2, 4] = [years, months, days],
        # and we use '\u200d' (zero width joiner) as placeholder.
        years=isinstance(years, int) and str(years) or '\u200d',
        year_str=isinstance(years, int) and year_str or '\u200d',
        months=isinstance(months, int) and str(months) or '\u200d',
        month_str=isinstance(months, int) and month_str or '\u200d',
        days=isinstance(days, int) and str(days) or '\u200d',
        day_str=isinstance(days, int) and day_str or '\u200d')


def get_age_for_comparison(age, type='y'):
    """ Get a age number used to conparison age."""
    (y, m, d) = parse_compute_age(age)

    if not isinstance(y, int):
        y = 0

    if not isinstance(m, int):
        m = 0

    if not isinstance(d, int):
        d = 0

    days = y * 365 + m * 30.5 + d
    months = days / 30.5
    years = days / 365

    if type == 'y':
        return years
    elif type == 'm':
        return months
    elif type == 'd':
        return days
    else:
        return years


def parse_compute_age(age):
    """ Parse age returned by compute_age_from_dates function."""
    if isinstance(age, str):
        return parse_compute_age_str(age)
    else:
        return age


def parse_compute_age_str(age_str):
    """ Parse age string returned by compute_age_from_dates function"""
    age_str = age_str.strip()
    sep = '\u200b'  # Zero width space
    if age_str.endswith(sep):
        return parse_compute_age_str_with_zero_width_space(age_str)
    else:
        return parse_compute_age_str_with_one_char_string(age_str)


def parse_compute_age_str_with_zero_width_space(age_str):
    """Parse age string seperate with zero width space.

    '10y 2m 03d' => [10, 2, 3]

    'y', 'm', 'd' can be translated to other languages, and this three
    string are surrounded with zero width space: '\u200b'.

    """
    sep = '\u200b'  # Zero width space
    age = age_str.split(sep)
    try:
        years = int(age[0])
    except BaseException:
        years = None
    try:
        months = int(age[2])
    except BaseException:
        months = None
    try:
        days = int(age[4])
    except BaseException:
        days = None
    return [years, months, days]


def parse_compute_age_str_with_one_char_string(age_str):
    """ Parse age string which is like: '10y 2m 03d'.

    'y', 'm', 'd' are 1 char strings, which can be translated to other
    languages.

    For example:

    '10y 2m 03d' => [10, 2, 3]

    Note: Previously, compute_age_from_dates will return '10y 4m 4d'
    style age string, so this function was retained for compatibility.
    """
    try:
        year = int(age_str.split(' ')[0][:-1])
        month = int(age_str.split(' ')[1][:-1])
        day = int(age_str.split(' ')[2][:-1])
        return [year, month, day]
    except BaseException:
        return [None, None, None]


def get_institution():
    # Retrieve the institution associated to this GNU Health instance
    # That is associated to the Company.
    pool = Pool()
    Company = pool.get('company.company')
    Institution = pool.get('gnuhealth.institution')
    company = Company.__table__()
    institution = Institution.__table__()

    company_id = Transaction().context.get('company')

    cursor = Transaction().connection.cursor()
    cursor.execute(*company.join(institution, condition=(
        institution.party == company.party)).select(
        institution.id,
        where=(company.id == company_id)))
    institution_id = cursor.fetchone()
    if institution_id:
        return int(institution_id[0])


def get_health_professional(required=True):
    # Get the professional associated to the internal user id
    # that logs into GNU Health
    # If the method is called with the arg "required" as False, then
    # the error message won't be shown in the case of not finding
    # the corresponding healthprof (eg, creating a new appointment)
    pool = Pool()
    Party = pool.get('party.party')
    Professional = pool.get('gnuhealth.healthprofessional')
    party = Party.__table__()
    professional = Professional.__table__()

    cursor = Transaction().connection.cursor()
    cursor.execute(
        *party.join(professional,
                    condition=(professional.party == party.id)).select(
            professional.id,
            where=(
                (party.is_healthprof)
                & (party.internal_user == Transaction().user))))
    healthprof_id = cursor.fetchone()
    if healthprof_id:
        return int(healthprof_id[0])
    else:
        if required and not pool.test:
            raise NoAssociatedHealthProfessional(gettext(
                ('health.msg_no_associated_health_professional'))
            )


# Matplotlib will be used by many report.py in the future, so we add a
# setup function to here.
def matplotlib_setup(matplotlab):
    matplotlibrc = os.path.join(matplotlab.get_configdir(), 'matplotlibrc')
    if os.path.exists(matplotlibrc):
        print(f'Matplotlib: RC file: {matplotlibrc} '
              'is found, just use it.')
    else:
        rc_conf_json = gettext('health.msg_matplotlib_rc_config_json_str')
        rc_conf = json.loads(rc_conf_json)
        rc_conf.pop('@comment', None)
        matplotlab.rcParams.update(rc_conf)
        print(f'Matplotlib: Use rcParams: {rc_conf}.')


def get_institution_currency():
    """ Retrieves this health insitution (company) currency.
    """
    Company = Pool().get('company.company')
    company_id = Transaction().context.get('company')
    company = Company(company_id)
    if (company.currency):
        return company.currency.id



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#                  exceptions.py: Exceptions classes                    #
#########################################################################

from trytond.exceptions import UserError
from trytond.model.exceptions import ValidationError


class WrongDateofBirth(ValidationError):
    pass


class DateHealedBeforeDx(ValidationError):
    pass


class EndTreatmentDateBeforeStart(ValidationError):
    pass


class MedEndDateBeforeStart(ValidationError):
    pass


class NextDoseBeforeFirst(ValidationError):
    pass


class DrugPregnancySafetyCheck(ValidationError):
    pass


class DrugAllergySafetyCheck(ValidationError):
    pass


class PrescriptionSafetyCheck(ValidationError):
    pass


class NoAssociatedHealthProfessional(ValidationError):
    pass


class EvaluationEndBeforeStart(ValidationError):
    pass


class MustBeAPerson(ValidationError):
    pass


class DupOfficialName(ValidationError):
    pass


class FedAccountMismatch(ValidationError):
    pass


class BirthCertDateMismatch(ValidationError):
    pass


class DupMainSP(ValidationError):
    pass


class NoAppointmentSelected(UserError):
    pass


class CanNotModifyVaccination(UserError):
    pass


class PatientDiseaseAlreadyExists(UserError):
    pass



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH REPORTING package                      #
#                       health_stock.py: main module                    #
#########################################################################
from datetime import datetime
from trytond.model import fields
from trytond.pyson import Eval, Not, Bool
from trytond.pool import Pool, PoolMeta
from trytond.i18n import gettext

from .exceptions import (ExpiredVaccine)

__all__ = ['Party', 'Lot', 'Move',
           'PatientPrescriptionOrder', 'PatientVaccination']

_STATES = {
    'readonly': Eval('state') == 'done',
}
_DEPENDS = ['state']


class Party(metaclass=PoolMeta):
    __name__ = 'party.party'
    warehouse = fields.Many2One(
        'stock.location', 'Warehouse',
        domain=[('type', 'in', ['warehouse', 'storage'])],
        states={
            'invisible': Not(Bool(Eval('is_pharmacy'))),
            'required': Bool(Eval('is_pharmacy')),
        },
        depends=['is_pharmacy'])

    @classmethod
    def default_warehouse(cls):
        Location = Pool().get('stock.location')
        locations = Location.search(cls.warehouse.domain)
        if len(locations) == 1:
            return locations[0].id


class Lot(metaclass=PoolMeta):
    __name__ = 'stock.lot'
    expiration_date = fields.Date('Expiration Date')


class Move(metaclass=PoolMeta):
    __name__ = 'stock.move'

    @classmethod
    def _get_origin(cls):
        return super(Move, cls)._get_origin() + [
            'gnuhealth.prescription.order',
            'gnuhealth.vaccination',
        ]


class PatientPrescriptionOrder(metaclass=PoolMeta):
    __name__ = 'gnuhealth.prescription.order'
    moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)

    @classmethod
    def __setup__(cls):
        super(PatientPrescriptionOrder, cls).__setup__()
        cls.pharmacy.states['readonly'] &= Bool(Eval('moves'))

    @classmethod
    def copy(cls, prescriptions, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['moves'] = None
        return super(PatientPrescriptionOrder, cls).copy(
            prescriptions, default=default)


class PatientVaccination(metaclass=PoolMeta):
    __name__ = 'gnuhealth.vaccination'
    moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
    location = fields.Many2One(
        'stock.location',
        'Stock Location', domain=[('type', '=', 'storage')])

    product = fields.Many2One('product.product', 'Product')

    lot = fields.Many2One(
        'stock.lot', 'Lot', depends=['product'],
        domain=[('product', '=', Eval('product'))],
        help="This field includes the lot number and expiration date")

    @fields.depends('lot', 'date')
    def on_change_lot(self):
        # Check expiration date on the vaccine lot
        if self.lot and self.date:
            if self.lot.expiration_date < datetime.date(self.date):
                raise ExpiredVaccine(
                    gettext('health_stock.msg_expired_vaccine')
                )

    @classmethod
    def copy(cls, vaccinations, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['moves'] = None
        return super(PatientVaccination, cls).copy(
            vaccinations, default=default)

    @fields.depends('vaccine')
    def on_change_vaccine(self):
        if self.vaccine:
            self.product = self.vaccine.product.id
        else:
            self.product = None



# SPDX-FileCopyrightText: 2020 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH ICU package                            #
#              __init__.py: Package declaration file                    #
#########################################################################

from trytond.pool import Pool
from . import health_icu


def register():
    Pool.register(
        health_icu.InpatientRegistration,
        health_icu.InpatientIcu,
        health_icu.Glasgow,
        health_icu.ApacheII,
        health_icu.MechanicalVentilation,
        health_icu.ChestDrainageAssessment,
        health_icu.PatientRounding,
        module='health_icu', type_='model')



# Copyright (C) 2008-2025 Luis Falcon <falcon@gnuhealth.org>
# Copyright (C) 2011-2025 GNU Solidario <health@gnusolidario.org>
# Copyright (C) 2013  Sebastian Marro <smarro@gnusolidario.org>
# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from trytond.wizard import Wizard, StateView, Button, StateTransition
from trytond.model import ModelView
from trytond.transaction import Transaction
from trytond.pool import Pool
from trytond.i18n import gettext
from trytond.modules.health.core import get_institution_currency

from ..exceptions import (StockMoveExists, NoStockOrigin)

__all__ = ['CreateVaccinationStockMoveInit', 'CreateVaccinationStockMove']


class CreateVaccinationStockMoveInit(ModelView):
    'Create Vaccination Stock Move Init'
    __name__ = 'gnuhealth.vaccination.stock.move.init'


class CreateVaccinationStockMove(Wizard):
    'Create Vaccination Stock Move'
    __name__ = 'gnuhealth.vaccination.stock.move.create'

    start = StateView(
        'gnuhealth.vaccination.stock.move.init',
        'health_stock.view_create_vaccination_stock_move', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Create Stock Move', 'create_stock_move', 'tryton-ok', True)
        ])
    create_stock_move = StateTransition()

    def transition_create_stock_move(self):
        pool = Pool()
        StockMove = pool.get('stock.move')
        Vaccination = pool.get('gnuhealth.vaccination')

        vaccinations = Vaccination.browse(Transaction().context.get(
            'active_ids'))
        for vaccination in vaccinations:

            if vaccination.moves:
                raise StockMoveExists(
                    gettext('health_stock.msg_stock_move_exists')
                )

            if not vaccination.location:
                raise NoStockOrigin(
                    gettext('health_stock.msg_no_location_origin')
                )

            lines = []

            line_data = {}
            line_data['origin'] = str(vaccination)
            line_data['from_location'] = \
                vaccination.location.id
            line_data['to_location'] = \
                vaccination.patient.party.customer_location.id
            line_data['product'] = \
                vaccination.vaccine.product.id
            line_data['unit_price'] = \
                vaccination.vaccine.product.list_price
            line_data['cost_price'] = \
                vaccination.vaccine.product.cost_price
            line_data['quantity'] = 1
            line_data['unit'] = \
                vaccination.vaccine.product.default_uom.id
            # Use the institution currency in the stock move
            line_data['currency'] = get_institution_currency()
            line_data['state'] = 'draft'
            lines.append(line_data)

            moves = StockMove.create(lines)

            StockMove.assign(moves)
            StockMove.do(moves)

        return 'end'
